import time
import logging

from svim.SVIM_intra import analyze_alignment_indel
from svim.SVIM_inter import analyze_read_segments
from svim.SVIM_COLLECT import retrieve_other_alignments


def span_position_distance(candidate, signature, position_distance_normalizer):
    if candidate.type == "INS" or candidate.type == "DUP_INT":
        c_contig, c_start, c_end = candidate.get_destination()
    else:
        c_contig, c_start, c_end = candidate.get_source()
    if signature.type == "DUP_INT":
        s_contig, s_start, s_end = signature.get_destination()
    else:
        s_contig, s_start, s_end = signature.get_source()
    #ins signatures can support dup_int candidates and vice versa
    if not (candidate.type == "INS" and signature.type == "DUP_INT") and \
       not (candidate.type == "DUP_INT" and signature.type == "INS") and \
       candidate.type != signature.type:
        return float("inf")
    if c_contig != s_contig:
        return float("inf")
    span1 = c_end - c_start
    span2 = s_end - s_start
    center1 = (c_start + c_end) // 2
    center2 = (s_start + s_end) // 2
    position_distance = min(abs(c_start - s_start), abs(c_end - s_end), abs(center1 - center2)) / position_distance_normalizer
    span_distance = abs(span1 - span2) / max(span1, span2)
    return position_distance + span_distance


def genotype(candidates, bam, type, options):
    num_candidates = len(candidates)
    for nr, candidate in enumerate(candidates):
        if (nr+1) % 10000 == 0:
            logging.info("Processed {0} of {1} candidates".format(nr+1, num_candidates))
        if candidate.score < options.minimum_score:
            continue
        #Fetch alignments around variant locus
        if type == "INS" or type == "DUP_INT":
            contig, start, end = candidate.get_destination()
            #We need the insertion locus on the reference for which end is equal to start
            end = start
        else:
            contig, start, end = candidate.get_source()
        contig_length = bam.get_reference_length(contig)
        alignment_it = bam.fetch(contig=contig, start=max(0, start-1000), stop=min(contig_length, end+1000))

        reads_supporting_variant = set([sig.read for sig in candidate.members])
        #Count reads that overlap the locus and therefore support the reference
        reads_supporting_reference = set()        
        #Loop through fetched alignments
        aln_no = 0
        while aln_no < 500:
            try:
                current_alignment = next(alignment_it)
            except StopIteration:
                break
            
            if current_alignment.query_name in reads_supporting_variant:
                continue
            if current_alignment.is_unmapped or current_alignment.is_secondary or current_alignment.mapping_quality < options.min_mapq:
                continue
            aln_no += 1

            if type == "DEL" or type == "INV":
                minimum_overlap = min((end - start) / 2, 2000)
                if (current_alignment.reference_start < (end - minimum_overlap) and current_alignment.reference_end > (end + 100) or
                    current_alignment.reference_start < (start - 100) and current_alignment.reference_end > (start + minimum_overlap)):
                    reads_supporting_reference.add(current_alignment.query_name)
            if type == "INS" or type == "DUP_INT":
                if current_alignment.reference_start < (start - 100) and current_alignment.reference_end > (end + 100):
                    reads_supporting_reference.add(current_alignment.query_name)

        if (len(reads_supporting_variant) + len(reads_supporting_reference)) >= options.minimum_depth:
            candidate.support_fraction = len(reads_supporting_variant) / (len(reads_supporting_variant) + len(reads_supporting_reference))
            if candidate.support_fraction >= options.homozygous_threshold:
                candidate.genotype = "1/1"
            elif candidate.support_fraction >= options.heterozygous_threshold and candidate.support_fraction < options.homozygous_threshold:
                candidate.genotype = "0/1"
            elif candidate.support_fraction < options.heterozygous_threshold:
                candidate.genotype = "0/0"
            else:
                candidate.genotype = "./."
        elif (len(reads_supporting_variant) + len(reads_supporting_reference)) > 0:
            candidate.support_fraction = len(reads_supporting_variant) / (len(reads_supporting_variant) + len(reads_supporting_reference))
            candidate.genotype = "./."
        else:
            candidate.support_fraction = "."
            candidate.genotype = "./."
        candidate.ref_reads = len(reads_supporting_reference)
        candidate.alt_reads = len(reads_supporting_variant)


from __future__ import print_function

import sys
from bisect import bisect_left
from collections import defaultdict
from math import pow, sqrt

from svim.SVSignature import SignatureTranslocation, SignatureInsertionFrom, SignatureClusterBiLocal
from svim.SVCandidate import CandidateDuplicationInterspersed
from svim.SVIM_clustering import span_position_distance_clusters

def flag_cutpaste_candidates(insertion_from_signature_clusters, deletion_signature_clusters, options):
    """Flag duplication signature clusters if they overlap a deletion"""
    int_duplication_candidates = []
    for ins_cluster in insertion_from_signature_clusters:
        # Compute distances of every deletion cluster to the current insertion/duplication
        distances = [(del_index, span_position_distance_clusters(del_cluster, ins_cluster, options.position_distance_normalizer)) \
                     for del_index, del_cluster in enumerate(deletion_signature_clusters)]
        closest_deletion_index, closest_deletion = sorted(distances, key=lambda obj: obj[1])[0]
        source_contig, source_start, source_end = ins_cluster.get_source()
        dest_contig, dest_start, dest_end = ins_cluster.get_destination()
        # If close deletion cluster found
        if closest_deletion <= options.del_ins_dup_max_distance:
            #Potential cut&paste insertion
            int_duplication_candidates.append(CandidateDuplicationInterspersed(source_contig, source_start, source_end, dest_contig, dest_start, dest_end, ins_cluster.members, ins_cluster.score, ins_cluster.std_span, ins_cluster.std_pos, cutpaste=True))
        else:
            #Interspersed duplication
            int_duplication_candidates.append(CandidateDuplicationInterspersed(source_contig, source_start, source_end, dest_contig, dest_start, dest_end, ins_cluster.members, ins_cluster.score, ins_cluster.std_span, ins_cluster.std_pos, cutpaste=False))
    return int_duplication_candidates


def get_closest_index(input_list, input_number):
    """
    Assumes input_list is sorted. Returns index of closest value to input_number.

    If two numbers are equally close, return the index of the smallest number.
    """
    if len(input_list) < 1:
        return None
    pos = bisect_left(input_list, input_number)
    if pos == 0:
        return 0
    if pos == len(input_list):
        return len(input_list) - 1
    before = input_list[pos - 1]
    after = input_list[pos]
    if after - input_number < input_number - before:
       return pos
    else:
       return pos - 1


def distance_positions(position1, position2):
    return float("inf") if position1[0] != position2[0] else abs(position1[1] - position2[1])


def calculate_score_insertion(main_score, translocation_distances, translocation_stds, destination_stds):
    """Calculate the score of a merged insertion or duplication detected from an insertion.
       Parameters: - main_score - score of the underlying main insertion
                   - translocation_distances - mean distance of the translocation clusters flanking the main insertion (left)
                   - translocation_stds - standard deviation of the translocation clusters flanking the main insertion (left)
                   - destination_stds - standard deviations of the left and right translocation destinations"""
    #scale translocation distance to [0, 1] range
    td0 = max(0, 100 - translocation_distances[0]) / 100
    td1 = max(0, 100 - translocation_distances[1]) / 100

    #scale translocation std to [0, 1] range
    if translocation_stds[0] == None:
        ts0 = 1
    else:
        ts0 = max(0, 100 - translocation_stds[0]) / 100
    if translocation_stds[1] == None:
        ts1 = 1
    else:
        ts1 = max(0, 100 - translocation_stds[1]) / 100

    #scale destination stds to [0, 1] range
    if destination_stds[0] == None:
        ds0 = 1
    else:
        ds0 = max(0, 100 - destination_stds[0]) / 100
    if destination_stds[1] == None:
        ds1 = 1
    else:
        ds1 = max(0, 100 - destination_stds[1]) / 100

    #calculate final score as product of components
    product = td0 * td1 * ts0 * ts1 * ds0 * ds1
    final_score = pow(product, 1/6) * main_score
    return final_score


def merge_translocations_at_insertions(translocation_signature_clusters, insertion_signature_clusters, options):
    if len(insertion_signature_clusters) == 0:
        return [], []
    
    #add reverse translocation signature clusters
    reversed_translocation_signature_clusters = []
    for cluster in translocation_signature_clusters:
        reversed_cluster = SignatureClusterBiLocal(cluster.dest_contig, cluster.dest_start, cluster.dest_end,
                                                   cluster.source_contig, cluster.source_start, cluster.source_end, 
                                                   cluster.score, cluster.size, cluster.members, cluster.type, cluster.std_pos, cluster.std_span)
        reversed_cluster.direction1 = 'fwd' if cluster.direction2 == 'rev' else 'rev'
        reversed_cluster.direction2 = 'fwd' if cluster.direction1 == 'rev' else 'rev'
        reversed_translocation_signature_clusters.append(reversed_cluster)
    translocation_signature_clusters.extend(reversed_translocation_signature_clusters)

    translocation_partitions_fwdfwd_dict = defaultdict(list)
    translocation_partitions_revrev_dict = defaultdict(list)
    for cluster in translocation_signature_clusters:
        if cluster.direction1 == 'fwd' and cluster.direction2 == 'fwd':
            translocation_partitions_fwdfwd_dict[cluster.source_contig].append(cluster)
        elif cluster.direction1 == 'rev' and cluster.direction2 == 'rev':
            translocation_partitions_revrev_dict[cluster.source_contig].append(cluster)
    for contig in translocation_partitions_fwdfwd_dict.keys():
        translocation_partitions_fwdfwd_dict[contig] = sorted(translocation_partitions_fwdfwd_dict[contig], key=lambda cluster: cluster.get_key())
    for contig in translocation_partitions_revrev_dict.keys():
        translocation_partitions_revrev_dict[contig] = sorted(translocation_partitions_revrev_dict[contig], key=lambda cluster: cluster.get_key())

    translocation_partition_means_fwdfwd_dict = {}
    translocation_partition_stds_fwdfwd_dict = {}
    for contig in translocation_partitions_fwdfwd_dict.keys():
        translocation_partition_means_fwdfwd_dict[contig] = [cluster.source_start for cluster in translocation_partitions_fwdfwd_dict[contig]]
        translocation_partition_stds_fwdfwd_dict[contig] = [cluster.std_span for cluster in translocation_partitions_fwdfwd_dict[contig]]
    translocation_partition_means_revrev_dict = {}
    translocation_partition_stds_revrev_dict = {}
    for contig in translocation_partitions_revrev_dict.keys():
        translocation_partition_means_revrev_dict[contig] = [cluster.source_start for cluster in translocation_partitions_revrev_dict[contig]]
        translocation_partition_stds_revrev_dict[contig] = [cluster.std_span for cluster in translocation_partitions_revrev_dict[contig]]

    inserted_regions_to_remove = []
    insertion_from_signature_clusters = []
    for insertion_index, ins_cluster in enumerate(insertion_signature_clusters):
        ins_contig, ins_start, ins_end = ins_cluster.get_source()
        try:
            closest_to_start_fwdfwd_index = get_closest_index(translocation_partition_means_fwdfwd_dict[ins_contig], ins_start)
            closest_to_start_fwdfwd_mean = translocation_partition_means_fwdfwd_dict[ins_contig][closest_to_start_fwdfwd_index]
            closest_to_start_revrev_index = get_closest_index(translocation_partition_means_revrev_dict[ins_contig], ins_start)
            closest_to_start_revrev_mean = translocation_partition_means_revrev_dict[ins_contig][closest_to_start_revrev_index]
        except KeyError:
            continue
        # if translocations found close to start of insertion
        if abs(closest_to_start_fwdfwd_mean - ins_start) <= options.trans_sv_max_distance and abs(closest_to_start_revrev_mean - ins_start) <= options.trans_sv_max_distance:
            destination_from_start_fwdfwd = (translocation_partitions_fwdfwd_dict[ins_contig][closest_to_start_fwdfwd_index].dest_contig, translocation_partitions_fwdfwd_dict[ins_contig][closest_to_start_fwdfwd_index].dest_start)
            destination_from_start_revrev = (translocation_partitions_revrev_dict[ins_contig][closest_to_start_revrev_index].dest_contig, translocation_partitions_revrev_dict[ins_contig][closest_to_start_revrev_index].dest_start)
            destination_from_start_fwdfwd_std = translocation_partitions_fwdfwd_dict[ins_contig][closest_to_start_fwdfwd_index].std_pos
            destination_from_start_revrev_std = translocation_partitions_revrev_dict[ins_contig][closest_to_start_revrev_index].std_pos
            # if the two destinations have the right distance
            distance = abs(destination_from_start_revrev[1] - destination_from_start_fwdfwd[1])
            if destination_from_start_revrev[0] == destination_from_start_fwdfwd[0] and 0.95 <= ((ins_end - ins_start + 1) / (distance + 1)) <= 1.1:
                members = ins_cluster.members + translocation_partitions_fwdfwd_dict[ins_contig][closest_to_start_fwdfwd_index].members + translocation_partitions_revrev_dict[ins_contig][closest_to_start_revrev_index].members
                score = calculate_score_insertion(ins_cluster.score, 
                                                  [abs(closest_to_start_fwdfwd_mean - ins_start), abs(closest_to_start_revrev_mean - ins_start)], 
                                                  [translocation_partition_stds_fwdfwd_dict[ins_contig][closest_to_start_fwdfwd_index], translocation_partition_stds_revrev_dict[ins_contig][closest_to_start_revrev_index]], 
                                                  [destination_from_start_fwdfwd_std, destination_from_start_revrev_std])
                insertion_from_signature_clusters.append(SignatureClusterBiLocal(destination_from_start_revrev[0], min(destination_from_start_revrev[1], destination_from_start_fwdfwd[1]), max(destination_from_start_revrev[1], destination_from_start_fwdfwd[1]), ins_contig, ins_start, ins_start + distance, score, len(members), members, "DUP_INT", ins_cluster.std_span, ins_cluster.std_pos))
                inserted_regions_to_remove.append(insertion_index)

    return insertion_from_signature_clusters, inserted_regions_to_remove



import os
import logging
import re

from collections import defaultdict
from math import pow, sqrt
import time
from statistics import mean, stdev
from pysam import FastaFile
from spoa import poa
from cpuinfo import get_cpu_info

from svim.SVIM_clustering import form_partitions, partition_and_cluster_candidates, calculate_score
from svim.SVCandidate import CandidateInversion, CandidateDuplicationTandem, CandidateDeletion, CandidateNovelInsertion, CandidateBreakend
from svim.SVIM_merging import flag_cutpaste_candidates, merge_translocations_at_insertions


def write_candidates(working_dir, candidates):
    int_duplication_candidates, inversion_candidates, tan_duplication_candidates, deletion_candidates, novel_insertion_candidates, breakend_candidates = candidates

    if not os.path.exists(working_dir + '/candidates'):
        os.mkdir(working_dir + '/candidates')
    deletion_candidate_output = open(working_dir + '/candidates/candidates_deletions.bed', 'w')
    inversion_candidate_output = open(working_dir + '/candidates/candidates_inversions.bed', 'w')
    tandem_duplication_candidate_source_output = open(working_dir + '/candidates/candidates_tan_duplications_source.bed', 'w')
    tandem_duplication_candidate_dest_output = open(working_dir + '/candidates/candidates_tan_duplications_dest.bed', 'w')
    interspersed_duplication_candidate_source_output = open(working_dir + '/candidates/candidates_int_duplications_source.bed', 'w')
    interspersed_duplication_candidate_dest_output = open(working_dir + '/candidates/candidates_int_duplications_dest.bed', 'w')
    novel_insertion_candidate_output = open(working_dir + '/candidates/candidates_novel_insertions.bed', 'w')
    breakend_candidate_output = open(working_dir + '/candidates/candidates_breakends.bed', 'w')

    for candidate in deletion_candidates:
        print(candidate.get_bed_entry(), file=deletion_candidate_output)
    for candidate in int_duplication_candidates:
        bed_entries = candidate.get_bed_entries()
        print(bed_entries[0], file=interspersed_duplication_candidate_source_output)
        print(bed_entries[1], file=interspersed_duplication_candidate_dest_output)
    for candidate in inversion_candidates:
        print(candidate.get_bed_entry(), file=inversion_candidate_output)
    for candidate in tan_duplication_candidates:
        bed_entries = candidate.get_bed_entries()
        print(bed_entries[0], file=tandem_duplication_candidate_source_output)
        print(bed_entries[1], file=tandem_duplication_candidate_dest_output)
    for candidate in novel_insertion_candidates:
        print(candidate.get_bed_entry(), file=novel_insertion_candidate_output)
    for candidate in breakend_candidates:
        bed_entries = candidate.get_bed_entries()
        print(bed_entries[0], file=breakend_candidate_output)
        print(bed_entries[1], file=breakend_candidate_output)

    deletion_candidate_output.close()
    inversion_candidate_output.close()
    interspersed_duplication_candidate_source_output.close()
    interspersed_duplication_candidate_dest_output.close()
    tandem_duplication_candidate_source_output.close()
    tandem_duplication_candidate_dest_output.close()
    novel_insertion_candidate_output.close()
    breakend_candidate_output.close()


def sorted_nicely(vcf_entries):
    """ Sort the given vcf entries (in the form ((contig, start, end), vcf_string, sv_type)) in the way that humans expect.
        e.g. chr10 comes after chr2
        Algorithm adapted from https://blog.codinghorror.com/sorting-for-humans-natural-sort-order/"""
    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [ convert(c) for c in re.split('([0-9]+)', key) ]
    tuple_key = lambda entry: ( alphanum_key(str(entry[0][0])), entry[0][1], entry[0][2] )
    return sorted(vcf_entries, key = tuple_key)


def write_final_vcf(int_duplication_candidates,
                    inversion_candidates, 
                    tandem_duplication_candidates, 
                    deletion_candidates, 
                    novel_insertion_candidates, 
                    breakend_candidates, 
                    version, 
                    contig_names, 
                    contig_lengths,
                    types_to_output,
                    options):
    vcf_output = open(options.working_dir + '/variants.vcf', 'w')

    # Write header lines
    print("##fileformat=VCFv4.2", file=vcf_output)
    print("##fileDate={0}".format(time.strftime("%Y-%m-%d|%I:%M:%S%p|%Z|%z")), file=vcf_output)
    print("##source=SVIM-v{0}".format(version), file=vcf_output)
    for contig_name, contig_length in zip(contig_names, contig_lengths):
        print("##contig=<ID={0},length={1}>".format(contig_name, contig_length), file=vcf_output)
    if "DEL" in types_to_output:
        print("##ALT=<ID=DEL,Description=\"Deletion\">", file=vcf_output)
    if "INV" in types_to_output:
        print("##ALT=<ID=INV,Description=\"Inversion\">", file=vcf_output)
    if (not options.tandem_duplications_as_insertions and "DUP:TANDEM" in types_to_output) or \
       (not options.interspersed_duplications_as_insertions and "DUP:INT" in types_to_output):
        print("##ALT=<ID=DUP,Description=\"Duplication\">", file=vcf_output)
    if not options.tandem_duplications_as_insertions and "DUP:TANDEM" in types_to_output:
        print("##ALT=<ID=DUP:TANDEM,Description=\"Tandem Duplication\">", file=vcf_output)
    if not options.interspersed_duplications_as_insertions and "DUP:INT" in types_to_output:
        print("##ALT=<ID=DUP:INT,Description=\"Interspersed Duplication\">", file=vcf_output)
    if "INS" in types_to_output:
        print("##ALT=<ID=INS,Description=\"Insertion\">", file=vcf_output)
    if "BND" in types_to_output:
        print("##ALT=<ID=BND,Description=\"Breakend\">", file=vcf_output)
    print("##INFO=<ID=SVTYPE,Number=1,Type=String,Description=\"Type of structural variant\">", file=vcf_output)
    print("##INFO=<ID=CUTPASTE,Number=0,Type=Flag,Description=\"Genomic origin of interspersed duplication seems to be deleted\">", file=vcf_output)
    print("##INFO=<ID=END,Number=1,Type=Integer,Description=\"End position of the variant described in this record\">", file=vcf_output)
    print("##INFO=<ID=SVLEN,Number=1,Type=Integer,Description=\"Difference in length between REF and ALT alleles\">", file=vcf_output)
    print("##INFO=<ID=SUPPORT,Number=1,Type=Integer,Description=\"Number of reads supporting this variant\">", file=vcf_output)
    print("##INFO=<ID=STD_SPAN,Number=1,Type=Float,Description=\"Standard deviation in span of merged SV signatures\">", file=vcf_output)
    print("##INFO=<ID=STD_POS,Number=1,Type=Float,Description=\"Standard deviation in position of merged SV signatures\">", file=vcf_output)
    print("##INFO=<ID=STD_POS1,Number=1,Type=Float,Description=\"Standard deviation of breakend 1 position\">", file=vcf_output)
    print("##INFO=<ID=STD_POS2,Number=1,Type=Float,Description=\"Standard deviation of breakend 2 position\">", file=vcf_output)
    if options.insertion_sequences:
        print("##INFO=<ID=SEQS,Number=.,Type=String,Description=\"Insertion sequences from all supporting reads\">", file=vcf_output)
    if options.read_names:
        print("##INFO=<ID=READS,Number=.,Type=String,Description=\"Names of all supporting reads\">", file=vcf_output)
    if options.zmws:
        print("##INFO=<ID=ZMWS,Number=1,Type=Integer,Description=\"Number of supporting ZMWs (PacBio only)\">", file=vcf_output)
    print("##FILTER=<ID=hom_ref,Description=\"Genotype is homozygous reference\">", file=vcf_output)
    print("##FILTER=<ID=not_fully_covered,Description=\"Tandem duplication is not fully covered by a single read\">", file=vcf_output)
    print("##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">", file=vcf_output)
    print("##FORMAT=<ID=DP,Number=1,Type=Integer,Description=\"Read depth\">", file=vcf_output)
    print("##FORMAT=<ID=AD,Number=R,Type=Integer,Description=\"Read depth for each allele\">", file=vcf_output)
    if not options.tandem_duplications_as_insertions and "DUP:TANDEM" in types_to_output:
        print("##FORMAT=<ID=CN,Number=1,Type=Integer,Description=\"Copy number of tandem duplication (e.g. 2 for one additional copy)\">", file=vcf_output)
    print("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t" + options.sample, file=vcf_output)

    # Open reference genome sequence file
    sequence_alleles = not options.symbolic_alleles
    if sequence_alleles:
        try:
            reference = FastaFile(options.genome)
        except ValueError:
            logging.warning("The given reference genome is missing an index file ({path}.fai). Sequence alleles cannot be retrieved.".format(options.genome))
            sequence_alleles = False
        except IOError:
            logging.warning("The given reference genome is missing ({path}). Sequence alleles cannot be retrieved.".format(options.genome))
            sequence_alleles = False
    else:
        reference = None

    # Prepare VCF entries depending on command-line parameters
    vcf_entries = []
    if "DEL" in types_to_output:
        for candidate in deletion_candidates:
            vcf_entries.append((candidate.get_source(), candidate.get_vcf_entry(sequence_alleles, reference, options.read_names, options.zmws), "DEL"))
    if "INV" in types_to_output:
        for candidate in inversion_candidates:
            vcf_entries.append((candidate.get_source(), candidate.get_vcf_entry(sequence_alleles, reference, options.read_names, options.zmws), "INV"))
    if "INS" in types_to_output:
        for candidate in novel_insertion_candidates:
            vcf_entries.append((candidate.get_destination(), candidate.get_vcf_entry(sequence_alleles, reference, options.insertion_sequences, options.read_names, options.zmws), "INS"))
    if options.tandem_duplications_as_insertions:
        if "INS" in types_to_output:
            for candidate in tandem_duplication_candidates:
                vcf_entries.append((candidate.get_destination(), candidate.get_vcf_entry_as_ins(sequence_alleles, reference, options.read_names, options.zmws), "INS"))
    else:
        if "DUP:TANDEM" in types_to_output:
            for candidate in tandem_duplication_candidates:
                vcf_entries.append((candidate.get_source(), candidate.get_vcf_entry_as_dup(options.read_names, options.zmws), "DUP_TANDEM"))
    if options.interspersed_duplications_as_insertions:
        if "INS" in types_to_output:
            for candidate in int_duplication_candidates:
                vcf_entries.append((candidate.get_destination(), candidate.get_vcf_entry_as_ins(sequence_alleles, reference, options.read_names, options.zmws), "INS"))
    else:
        if "DUP:INT" in types_to_output:
            for candidate in int_duplication_candidates:
                vcf_entries.append((candidate.get_source(), candidate.get_vcf_entry_as_dup(options.read_names, options.zmws), "DUP_INT"))
    if "BND" in types_to_output:
        for candidate in breakend_candidates:
            vcf_entries.append(((candidate.get_source()[0], candidate.get_source()[1], candidate.get_source()[1] + 1), candidate.get_vcf_entry(options.read_names, options.zmws), "BND"))
            vcf_entries.append(((candidate.get_destination()[0], candidate.get_destination()[1], candidate.get_destination()[1] + 1), candidate.get_vcf_entry_reverse(options.read_names, options.zmws), "BND"))

    if sequence_alleles:
        reference.close()

    # Sort and write entries to VCF
    svtype_counter = defaultdict(int)
    for source, entry, svtype in sorted_nicely(vcf_entries):
        variant_id = "svim.{svtype}.{number}".format(svtype = svtype, number = svtype_counter[svtype] + 1)
        entry_with_id = entry.replace("PLACEHOLDERFORID", variant_id, 1)
        svtype_counter[svtype] += 1
        print(entry_with_id, file=vcf_output)

    vcf_output.close()

def generate_insertion_consensus(ins_cluster, reference, window_padding = 100, maximum_haplotype_length = 10000, allowed_size_deviation = 2.0):
    #compute window containing all members of cluster
    member_pos = [member.start for member in ins_cluster.members]
    window_start = min(member_pos) - window_padding
    window_end = max(member_pos) + window_padding

    #construct haplotype sequences from all reads
    haplotypes = []
    for member in ins_cluster.members:
        haplotype_sequence = reference.fetch(ins_cluster.contig, max(0, window_start), max(0, member.start)).upper()
        haplotype_sequence += member.sequence.upper()
        haplotype_sequence += reference.fetch(ins_cluster.contig, max(0, member.start), max(0, window_end)).upper()
        haplotypes.append(haplotype_sequence)
    largest_haplotype_length = max([len(h) for h in haplotypes])
    if largest_haplotype_length > maximum_haplotype_length:
        logging.info("Skipping consensus computation for insertion with haplotypes exceeding maximum length ({0} > {1})".format(largest_haplotype_length, maximum_haplotype_length))
        return (1, ())

    #compute consensus of all reads using SPOA
    try:
        consensus_reads, msa_reads = poa(haplotypes, algorithm=1, m=2, n=-4, g=-4, e=-2, q=-24, c=-1)
    except MemoryError:
        logging.warning("Error: spoa ran out of memory while computing a consensus sequence for a cluster of insertion signatures (size = {0}, maximum haplotype length = {1}).".format(ins_cluster.size, largest_haplotype_length))
        return (2, ())
    except ValueError as ve:
        logging.warning("Error: spoa failed while computing a consensus sequence for a cluster of insertion signatures (size = {0}, maximum haplotype length = {1}): {2}".format(ins_cluster.size, largest_haplotype_length, ve))
        return (2, ())

    #re-align consensus sequence to reference sequence in the window
    ref_sequence = reference.fetch(ins_cluster.contig, max(0, window_start), max(0, window_end)).upper()
    try:
        consensus_reads_ref, msa_reads_ref = poa([consensus_reads, ref_sequence], algorithm=1, m=2, n=-4, g=-4, e=-2, q=-24, c=-1)
    except MemoryError:
        logging.warning("Error: spoa ran out of memory while aligning the consensus sequence for a cluster of insertion signatures (size = {0}, maximum haplotype length = {1}).".format(ins_cluster.size, largest_haplotype_length))
        return (2, ())
    except ValueError as ve:
        logging.warning("Error: spoa failed while aligning the consensus sequence for a cluster of insertion signatures (size = {0}, maximum haplotype length = {1}): {2}".format(ins_cluster.size, largest_haplotype_length, ve))
        return (2, ())

    #locate insertion relative to reference and check whether size is close to expected size
    expected_size = ins_cluster.end - ins_cluster.start
    matches = []
    for match in re.finditer(r'-+', msa_reads_ref[1]):
        match_size = match.end() - match.start()
        size_ratio = max(match_size, expected_size) / min(match_size, expected_size)
        matches.append((match.start(), match_size, size_ratio))
    good_matches = [m for m in matches if m[2] < allowed_size_deviation]
    if len(good_matches) == 0:
        logging.info("Consensus failure (no suitable insertion found in realignment step). \
                      Expected size: {0}; Match sizes: {1}; Ref: {2}; Consensus: {3}".format(expected_size,
                                                                                             "/".join([str(m[1]) for m in matches]),
                                                                                             msa_reads_ref[1],
                                                                                             msa_reads_ref[0]))
        return (3, ())
    elif len(good_matches) == 1:
        realigned_insertion_start = max(0, window_start) + good_matches[0][0]
        realigned_insertion_size = good_matches[0][1]
        insertion_consensus = msa_reads_ref[0][good_matches[0][0]:good_matches[0][0]+good_matches[0][1]]
        logging.debug("Consensus success. Expected size: {0}; Consensus size: {1}; Consensus insertion: {2}".format(expected_size, realigned_insertion_size, insertion_consensus))
        return (0, (realigned_insertion_start, realigned_insertion_size, insertion_consensus))
    else:
        logging.info("Consensus failure (multiple suitable insertions found in realignment step). \
                      Expected size: {0}; Match sizes: {1}; Ref: {2}; Consensus: {3}".format(expected_size,
                                                                                             "/".join([str(m[1]) for m in matches]),
                                                                                             msa_reads_ref[1],
                                                                                             msa_reads_ref[0]))
        return (4, ())


def prepare_insertion_candidates(insertion_signature_clusters, options):
    novel_insertion_candidates = []
    if options.skip_consensus or ("sse4_1" not in get_cpu_info()["flags"]):
        if options.skip_consensus:
            logging.info("Skipping computation of insertion consensus sequences because of --skip_consensus flag.")
        else:
            logging.warning("Skipping computation of insertion consensus sequences "
                            "because CPU does not support SSE 4.1 instruction set.")
        for ins_cluster in insertion_signature_clusters:
            if ins_cluster.score > 0:
                novel_insertion_candidates.append(CandidateNovelInsertion(ins_cluster.contig,
                                                                              ins_cluster.start,
                                                                              ins_cluster.end,
                                                                              "",
                                                                              ins_cluster.members,
                                                                              ins_cluster.score,
                                                                              ins_cluster.std_span,
                                                                              ins_cluster.std_pos))
        return novel_insertion_candidates
    logging.info("Generating and realigning consensus sequence for insertions..")
    with FastaFile(options.genome) as reference:
        # 0-successful, 1-skipped, 2-failed, 3-no consensus, 4-multiple consensuses
        status_counter = [0, 0, 0, 0, 0]
        for ins_cluster in insertion_signature_clusters:
            if ins_cluster.score > 0:
                if len(ins_cluster.members) < 3:
                    novel_insertion_candidates.append(CandidateNovelInsertion(ins_cluster.contig,
                                                                              ins_cluster.start,
                                                                              ins_cluster.end,
                                                                              ins_cluster.members[0].sequence,
                                                                              ins_cluster.members,
                                                                              ins_cluster.score,
                                                                              ins_cluster.std_span,
                                                                              ins_cluster.std_pos))
                    continue
                status, consensus_result = generate_insertion_consensus(ins_cluster,
                                                                        reference,
                                                                        maximum_haplotype_length=options.max_consensus_length)
                try:
                    status_counter[status] += 1
                except KeyError:
                    logging.error("Unknown status {0} returned by consensus computation.".format(status))
                    continue
                # Successful
                if status == 0:
                    realigned_insertion_start, realigned_insertion_size, insertion_consensus = consensus_result
                    novel_insertion_candidates.append(CandidateNovelInsertion(ins_cluster.contig,
                                                                              realigned_insertion_start,
                                                                              realigned_insertion_start + realigned_insertion_size,
                                                                              insertion_consensus,
                                                                              ins_cluster.members,
                                                                              ins_cluster.score,
                                                                              ins_cluster.std_span,
                                                                              ins_cluster.std_pos))
                #Unsuccessful
                else:
                    novel_insertion_candidates.append(CandidateNovelInsertion(ins_cluster.contig,
                                                                              ins_cluster.start,
                                                                              ins_cluster.end,
                                                                              "",
                                                                              ins_cluster.members,
                                                                              ins_cluster.score,
                                                                              ins_cluster.std_span,
                                                                              ins_cluster.std_pos))
    message = ("Generated and realigned consensus sequences for {0} insertions "
               "({1} skipped, {2} failed with an error, {3} failed with no "
               "consensus, {4} failed with multiple consensuses).")
    logging.info(message.format(status_counter[0],
                                status_counter[1],
                                status_counter[2],
                                status_counter[3],
                                status_counter[4]))
    return novel_insertion_candidates


def combine_clusters(signature_clusters, options):
    deletion_signature_clusters, insertion_signature_clusters, inversion_signature_clusters, tandem_duplication_signature_clusters, insertion_from_signature_clusters, translocation_signature_clusters = signature_clusters

    ###############################
    # Create inversion candidates #
    ###############################
    inversion_candidates = []
    for inv_cluster in inversion_signature_clusters:
        inversion_candidates.append(CandidateInversion(inv_cluster.contig, inv_cluster.start, inv_cluster.end, inv_cluster.members, inv_cluster.score, inv_cluster.std_span, inv_cluster.std_pos))

    ########################################
    # Create tandem duplication candidates #
    ########################################
    tan_dup_candidates = []
    for tan_dup_cluster in tandem_duplication_signature_clusters:
        source_contig, source_start, source_end = tan_dup_cluster.get_source()
        dest_contig, dest_start, dest_end = tan_dup_cluster.get_destination()
        num_copies = int(round((dest_end - dest_start) / (source_end - source_start)))
        fully_covered = True if sum([sig.fully_covered for sig in tan_dup_cluster.members]) else False
        tan_dup_candidates.append(CandidateDuplicationTandem(tan_dup_cluster.source_contig, tan_dup_cluster.source_start, tan_dup_cluster.source_end, num_copies, fully_covered, tan_dup_cluster.members, tan_dup_cluster.score, tan_dup_cluster.std_span, tan_dup_cluster.std_pos))

    #####################################
    # Cluster translocation breakpoints #
    #####################################

    # Cluster translocations by contig and pos1
    # logging.info("Cluster translocation breakpoints..")
    # translocations_fwdfwd = [tra for tra in translocation_signature_clusters if tra.direction1 == "fwd" and tra.direction2 == "fwd"]
    # translocations_revrev = [tra for tra in translocation_signature_clusters if tra.direction1 == "rev" and tra.direction2 == "rev"]
    # translocations_fwdrev = [tra for tra in translocation_signature_clusters if tra.direction1 == "fwd" and tra.direction2 == "rev"]
    # translocations_revfwd = [tra for tra in translocation_signature_clusters if tra.direction1 == "rev" and tra.direction2 == "fwd"]
    # translocation_partitions_fwdfwd = form_partitions(translocations_fwdfwd, options.trans_partition_max_distance)
    # translocation_partitions_revrev = form_partitions(translocations_revrev, options.trans_partition_max_distance)
    # translocation_partitions_fwdrev = form_partitions(translocations_fwdrev, options.trans_partition_max_distance)
    # translocation_partitions_revfwd = form_partitions(translocations_revfwd, options.trans_partition_max_distance)

    ##############################
    # Create breakend candidates #
    ##############################

    breakend_candidates = []
    for tra_cluster in translocation_signature_clusters:
        breakend_candidates.append(CandidateBreakend(tra_cluster.source_contig, 
                                                     tra_cluster.source_start, 
                                                     tra_cluster.direction1, 
                                                     tra_cluster.dest_contig, 
                                                     tra_cluster.dest_start, 
                                                     tra_cluster.direction2, 
                                                     tra_cluster.members, 
                                                     tra_cluster.score, 
                                                     tra_cluster.std_span, 
                                                     tra_cluster.std_pos))

    ###################################################
    # Merge translocation breakpoints with insertions #
    ###################################################

    logging.info("Combine inserted regions with translocation breakpoints..")
    new_insertion_from_clusters, inserted_regions_to_remove_1 = merge_translocations_at_insertions(translocation_signature_clusters, insertion_signature_clusters, options)
    insertion_from_signature_clusters.extend(new_insertion_from_clusters)

    ############################################################################
    # Create interspersed duplication candidates and flag cut&paste insertions #
    ############################################################################

    logging.info("Create interspersed duplication candidates and flag cut&paste insertions..")
    int_duplication_candidates = flag_cutpaste_candidates(insertion_from_signature_clusters, deletion_signature_clusters, options)

    ###################################
    # Remove inserted region clusters #
    ###################################

    #find all inserted regions overlapping interspersed duplication or tandem duplication candidates
    int_duplication_iterator = iter(sorted(int_duplication_candidates, key=lambda cand: cand.get_destination()))
    tan_duplication_iterator = iter(sorted(tan_dup_candidates, key=lambda cand: cand.get_destination()))
    int_duplications_end = False
    tan_duplications_end = False
    inserted_regions_to_remove_2 = []

    try:
        current_int_duplication = next(int_duplication_iterator)
    except StopIteration:
        int_duplications_end = True

    try:
        current_tan_duplication = next(tan_duplication_iterator)
    except StopIteration:
        tan_duplications_end = True

    for inserted_region_index, inserted_region in enumerate(insertion_signature_clusters):
        contig1, start1, end1 = inserted_region.get_source()
        length1 = end1 - start1
        if not int_duplications_end:
            contig2, start2, end2 = current_int_duplication.get_destination()
            while contig2 < contig1 or (contig2 == contig1 and end2 < start1):
                try:
                    current_int_duplication = next(int_duplication_iterator)
                    contig2, start2, end2 = current_int_duplication.get_destination()
                except StopIteration:
                    int_duplications_end = True
                    break
        if not int_duplications_end:
            length2 = end2 - start2
            #if overlapping interspersed duplication of similar length
            if contig2 == contig1 and start2 < end1 and (length1 - length2) / max(length1, length2) < 0.2:
                inserted_regions_to_remove_2.append(inserted_region_index)
        else:
            if not tan_duplications_end:
                contig2, start2, end2 = current_tan_duplication.get_destination()
                while contig2 < contig1 or (contig2 == contig1 and end2 < start1):
                    try:
                        current_tan_duplication = next(tan_duplication_iterator)
                        contig2, start2, end2 = current_tan_duplication.get_destination()
                    except StopIteration:
                        tan_duplications_end = True
                        break
            if not tan_duplications_end:
                length2 = end2 - start2
                #if overlapping tandem duplication of similar length
                if contig2 == contig1 and start2 < end1 and (length1 - length2) / max(length1, length2) < 0.2:
                    inserted_regions_to_remove_2.append(inserted_region_index)

    # remove found inserted regions
    all_inserted_regions_to_remove = sorted(list(set(inserted_regions_to_remove_1 + inserted_regions_to_remove_2)), reverse=True)
    for ins_index in all_inserted_regions_to_remove:
        del(insertion_signature_clusters[ins_index])

    ##############################
    # Create deletion candidates #
    ##############################
    deletion_candidates = []
    for del_cluster in deletion_signature_clusters:
        if del_cluster.score > 0:
            deletion_candidates.append(CandidateDeletion(del_cluster.contig, del_cluster.start, del_cluster.end, del_cluster.members, del_cluster.score, del_cluster.std_span, del_cluster.std_pos))

    #####################################
    # Create novel insertion candidates #
    #####################################
    novel_insertion_candidates = prepare_insertion_candidates(insertion_signature_clusters, options)

    ######################
    # Cluster candidates #
    ######################
    logging.info("Cluster interspersed duplication candidates one more time..")
    final_int_duplication_candidates = partition_and_cluster_candidates(int_duplication_candidates, options, "interspersed duplication candidates")

    return (deletion_candidates, inversion_candidates, final_int_duplication_candidates, tan_dup_candidates, novel_insertion_candidates, breakend_candidates)



class Candidate:
    """Candidate class for structural variant candidates. Candidates reflect the final SV types and can be merged from signatures of several reads.
    """
    def __init__(self, source_contig, source_start, source_end, members, score, std_span, std_pos, support_fraction = ".", genotype = "./.", ref_reads = None, alt_reads = None):
        self.source_contig = source_contig
        self.source_start = source_start
        self.source_end = source_end
        
        self.members = members
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.type = None
        self.support_fraction = support_fraction
        self.genotype = genotype
        self.ref_reads = ref_reads
        self.alt_reads = alt_reads


    def get_source(self):
        return (self.source_contig, self.source_start, self.source_end)


    def get_key(self):
        contig, start, end = self.get_source()
        return (self.type, contig, end)


    def downstream_distance_to(self, candidate2):
        """Return distance >= 0 between this candidate's end and the start of candidate2."""
        this_contig, this_start, this_end = self.get_source()
        other_contig, other_start, other_end = candidate2.get_source()
        if self.type == candidate2.type and this_contig == other_contig:
            return max(0, other_start - this_end)
        else:
            return float("inf")


    def get_std_span(self, ndigits=2):
        if self.std_span:
            return round(self.std_span, ndigits)
        else:
            return "."


    def get_std_pos(self, ndigits=2):
        if self.std_pos:
            return round(self.std_pos, ndigits)
        else:
            return "."

    def get_bed_entry(self):
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}".format(self.source_contig, self.source_start, self.source_end, "{0};{1};{2}".format(self.type, self.get_std_span(), self.get_std_pos()), self.score, ".", "["+"][".join([ev.as_string("|") for ev in self.members])+"]")


    def get_vcf_entry(self):
        raise NotImplementedError


class CandidateDeletion(Candidate):
    def __init__(self, source_contig, source_start, source_end, members, score, std_span, std_pos, support_fraction = ".", genotype = "./.", ref_reads = None, alt_reads = None):
        self.source_contig = source_contig
        #0-based start of the deletion (first deleted base)
        self.source_start = max(0, source_start)
        #0-based end of the deletion (one past the last deleted base)
        self.source_end = source_end

        self.members = members
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.type = "DEL"
        self.support_fraction = support_fraction
        self.genotype = genotype
        self.ref_reads = ref_reads
        self.alt_reads = alt_reads


    def get_vcf_entry(self, sequence_alleles = False, reference = None, read_names = False, zmws = False):
        contig, start, end = self.get_source()
        if self.ref_reads != None and self.alt_reads != None:
            dp_string = str(self.ref_reads + self.alt_reads)
        else:
            dp_string = "."
        filters = []
        if self.genotype == "0/0":
            filters.append("hom_ref")
        if sequence_alleles:
            ref_allele = reference.fetch(contig, max(0, start-1), end).upper()
            alt_allele = reference.fetch(contig, max(0, start-1), start).upper()
        else:
            ref_allele = "N"
            alt_allele = "<" + self.type + ">"
        info_template="SVTYPE={0};END={1};SVLEN={2};SUPPORT={3};STD_SPAN={4};STD_POS={5}"
        info_string = info_template.format(self.type, 
                                           end, 
                                           start - end, 
                                           len(set([sig.read for sig in self.members])), 
                                           self.get_std_span(), 
                                           self.get_std_pos())
        read_ids = [member.read for member in self.members]
        if read_names:
            info_string += ";READS={0}".format(",".join(read_ids))
        if zmws:
            valid_pacbio_names = True
            zmw_list = set()
            for read_id in read_ids:
                fields = read_id.split("/")
                if len(fields) != 3:
                    valid_pacbio_names = False
                    break
                zmw_list.add("/".join(fields[0:2]))
            if valid_pacbio_names:
                info_string += ";ZMWS={0}".format(len(zmw_list))
        return "{chrom}\t{pos}\t{id}\t{ref}\t{alt}\t{qual}\t{filter}\t{info}\t{format}\t{samples}".format(
                    chrom=contig,
                    pos=max(1, start),
                    id="PLACEHOLDERFORID",
                    ref=ref_allele,
                    alt=alt_allele,
                    qual=int(self.score),
                    filter="PASS" if len(filters) == 0 else ";".join(filters),
                    info=info_string,
                    format="GT:DP:AD",
                    samples="{gt}:{dp}:{ref},{alt}".format(gt=self.genotype, dp=dp_string, ref=self.ref_reads if self.ref_reads != None else ".", alt=self.alt_reads if self.alt_reads != None else "."))


class CandidateInversion(Candidate):
    def __init__(self, source_contig, source_start, source_end, members, score, std_span, std_pos, support_fraction = ".", genotype = "./.", ref_reads = None, alt_reads = None):
        self.source_contig = source_contig
        #0-based start of the inversion (first inverted base)
        self.source_start = max(0, source_start)
        #0-based end of the inversion (one past the last inverted base)
        self.source_end = source_end

        self.members = members
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.type = "INV"
        self.support_fraction = support_fraction
        self.genotype = genotype
        self.ref_reads = ref_reads
        self.alt_reads = alt_reads

        self.complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'}


    def get_vcf_entry(self, sequence_alleles = False, reference = None, read_names = False, zmws = False):
        contig, start, end = self.get_source()
        if self.ref_reads != None and self.alt_reads != None:
            dp_string = str(self.ref_reads + self.alt_reads)
        else:
            dp_string = "."
        filters = []
        if self.genotype == "0/0":
            filters.append("hom_ref")
        if sequence_alleles:
            ref_allele = reference.fetch(contig, start, end).upper()
            alt_allele = "".join(self.complement.get(base.upper(), base.upper()) for base in reversed(ref_allele))
        else:
            ref_allele = "N"
            alt_allele = "<" + self.type + ">"
        info_template="SVTYPE={0};END={1};SUPPORT={2};STD_SPAN={3};STD_POS={4}"
        info_string = info_template.format(self.type, 
                                            end, 
                                            len(set([sig.read for sig in self.members])), 
                                            self.get_std_span(), 
                                            self.get_std_pos())
        read_ids = [member.read for member in self.members]
        if read_names:
            info_string += ";READS={0}".format(",".join(read_ids))
        if zmws:
            valid_pacbio_names = True
            zmw_list = set()
            for read_id in read_ids:
                fields = read_id.split("/")
                if len(fields) != 3:
                    valid_pacbio_names = False
                    break
                zmw_list.add("/".join(fields[0:2]))
            if valid_pacbio_names:
                info_string += ";ZMWS={0}".format(len(zmw_list))
        return "{chrom}\t{pos}\t{id}\t{ref}\t{alt}\t{qual}\t{filter}\t{info}\t{format}\t{samples}".format(
                    chrom=contig,
                    pos=start+1,
                    id="PLACEHOLDERFORID",
                    ref=ref_allele,
                    alt=alt_allele,
                    qual=int(self.score),
                    filter="PASS" if len(filters) == 0 else ";".join(filters),
                    info=info_string,
                    format="GT:DP:AD",
                    samples="{gt}:{dp}:{ref},{alt}".format(gt=self.genotype, dp=dp_string, ref=self.ref_reads if self.ref_reads != None else ".", alt=self.alt_reads if self.alt_reads != None else "."))


class CandidateNovelInsertion(Candidate):
    def __init__(self, dest_contig, dest_start, dest_end, sequence, members, score, std_span, std_pos, support_fraction = ".", genotype = "./.", ref_reads = None, alt_reads = None):
        self.dest_contig = dest_contig
        #0-based start of the insertion (base after the insertion)
        self.dest_start = max(0, dest_start)
        #0-based start of the insertion (base after the insertion) + length of the insertion
        self.dest_end = dest_end
        self.sequence = sequence

        self.members = members
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.type = "INS"
        self.support_fraction = support_fraction
        self.genotype = genotype
        self.ref_reads = ref_reads
        self.alt_reads = alt_reads

    def get_destination(self):
        return (self.dest_contig, self.dest_start, self.dest_end)

    def get_bed_entry(self):
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}".format(self.dest_contig, self.dest_start, self.dest_end, "{0};{1};{2}".format(self.type, self.get_std_span(), self.get_std_pos()), self.score, ".", "["+"][".join([ev.as_string("|") for ev in self.members])+"]")

    def get_vcf_entry(self, sequence_alleles = False, reference = None, insertion_sequences = False, read_names = False, zmws = False):
        contig, start, end = self.get_destination()
        if self.ref_reads != None and self.alt_reads != None:
            dp_string = str(self.ref_reads + self.alt_reads)
        else:
            dp_string = "."
        filters = []
        if self.genotype == "0/0":
            filters.append("hom_ref")
        if sequence_alleles and self.sequence != "":
            ref_allele = reference.fetch(contig, max(0, start-1), max(0, start-1) + 1).upper()
            alt_allele = ref_allele + self.sequence
        else:
            ref_allele = "N"
            alt_allele = "<" + self.type + ">"
        info_template="SVTYPE={0};END={1};SVLEN={2};SUPPORT={3};STD_SPAN={4};STD_POS={5}"
        info_string = info_template.format(self.type, 
                                           start, 
                                           end - start, 
                                           len(set([sig.read for sig in self.members])), 
                                           self.get_std_span(), 
                                           self.get_std_pos()) 
        if insertion_sequences:
            insertion_seqs = [member.sequence for member in self.members]
            info_string += ";SEQS={0}".format(",".join(insertion_seqs))
        read_ids = [member.read for member in self.members]
        if read_names:
            info_string += ";READS={0}".format(",".join(read_ids))
        if zmws:
            valid_pacbio_names = True
            zmw_list = set()
            for read_id in read_ids:
                fields = read_id.split("/")
                if len(fields) != 3:
                    valid_pacbio_names = False
                    break
                zmw_list.add("/".join(fields[0:2]))
            if valid_pacbio_names:
                info_string += ";ZMWS={0}".format(len(zmw_list))
        return "{chrom}\t{pos}\t{id}\t{ref}\t{alt}\t{qual}\t{filter}\t{info}\t{format}\t{samples}".format(
                    chrom=contig,
                    pos=max(1, start),
                    id="PLACEHOLDERFORID",
                    ref=ref_allele,
                    alt=alt_allele,
                    qual=int(self.score),
                    filter="PASS" if len(filters) == 0 else ";".join(filters),
                    info=info_string,
                    format="GT:DP:AD",
                    samples="{gt}:{dp}:{ref},{alt}".format(gt=self.genotype, dp=dp_string, ref=self.ref_reads if self.ref_reads != None else ".", alt=self.alt_reads if self.alt_reads != None else "."))


class CandidateDuplicationTandem(Candidate):
    def __init__(self, source_contig, source_start, source_end, copies, fully_covered, members, score, std_span, std_pos, support_fraction = ".", genotype = "./.", ref_reads = None, alt_reads = None):
        self.source_contig = source_contig
        #0-based start of the region (first copied base)
        self.source_start = max(0, source_start)
        #0-based end of the region (one past the last copied base)
        self.source_end = source_end
        
        #number of copies inserted after end of the region (in tandem)
        self.copies = copies

        self.members = members
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.type = "DUP_TAN"
        self.support_fraction = support_fraction
        self.genotype = genotype
        self.ref_reads = ref_reads
        self.alt_reads = alt_reads
        self.fully_covered = fully_covered


    def get_destination(self):
        source_contig, source_start, source_end = self.get_source()
        return (source_contig, source_end, source_end + self.copies * (source_end - source_start))


    def get_bed_entries(self, sep="\t"):
        source_contig, source_start, source_end = self.get_source()
        dest_contig, dest_start, dest_end = self.get_destination()
        source_entry = sep.join(["{0}", "{1}", "{2}", "{3}", "{4}", "{5}", "{6}"]).format(source_contig, source_start,
                                                                                     source_end,
                                                                                     "tan_dup_source;>{0}:{1}-{2};{3};{4}".format(
                                                                                         dest_contig, dest_start,
                                                                                         dest_end, self.get_std_span(), self.get_std_pos()), self.score, ".",
                                                                                     "[" + "][".join(
                                                                                         [ev.as_string("|") for ev in
                                                                                          self.members]) + "]")
        dest_entry = sep.join(["{0}", "{1}", "{2}", "{3}", "{4}", "{5}", "{6}"]).format(dest_contig, dest_start, dest_end,
                                                                                   "tan_dup_dest;<{0}:{1}-{2};{3};{4}".format(
                                                                                       source_contig, source_start,
                                                                                       source_end, self.get_std_span(), self.get_std_pos()), self.score, ".",
                                                                                   "[" + "][".join(
                                                                                       [ev.as_string("|") for ev in
                                                                                        self.members]) + "]")
        return (source_entry, dest_entry)


    def get_vcf_entry_as_ins(self, sequence_alleles = False, reference = None, read_names = False, zmws = False):
        source_contig, source_start, source_end = self.get_source()
        dest_contig, dest_start, dest_end = self.get_destination()
        svtype = "INS"
        if self.ref_reads != None and self.alt_reads != None:
            dp_string = str(self.ref_reads + self.alt_reads)
        else:
            dp_string = "."
        filters = []
        if self.genotype == "0/0":
            filters.append("hom_ref")
        if not self.fully_covered:
            filters.append("not_fully_covered")
        if sequence_alleles:
            ref_allele = reference.fetch(source_contig, source_start, source_end).upper()
            alt_allele = ref_allele * (self.copies + 1)
        else:
            ref_allele = "N"
            alt_allele = "<" + self.type + ">"
        info_template="SVTYPE={0};END={1};SVLEN={2};SUPPORT={3};STD_SPAN={4};STD_POS={5}"
        info_string = info_template.format(svtype,
                                           source_end,
                                           dest_end - dest_start,
                                           len(set([sig.read for sig in self.members])),
                                           self.get_std_span(),
                                           self.get_std_pos())
        read_ids = [member.read for member in self.members]
        if read_names:
            info_string += ";READS={0}".format(",".join(read_ids))
        if zmws:
            valid_pacbio_names = True
            zmw_list = set()
            for read_id in read_ids:
                fields = read_id.split("/")
                if len(fields) != 3:
                    valid_pacbio_names = False
                    break
                zmw_list.add("/".join(fields[0:2]))
            if valid_pacbio_names:
                info_string += ";ZMWS={0}".format(len(zmw_list))
        return "{chrom}\t{pos}\t{id}\t{ref}\t{alt}\t{qual}\t{filter}\t{info}\t{format}\t{samples}".format(
                    chrom=source_contig,
                    pos=max(1, dest_start),
                    id="PLACEHOLDERFORID",
                    ref=ref_allele,
                    alt=alt_allele,
                    qual=int(self.score),
                    filter="PASS" if len(filters) == 0 else ";".join(filters),
                    info=info_string,
                    format="GT:DP:AD",
                    samples="{gt}:{dp}:{ref},{alt}".format(gt=self.genotype, dp=dp_string, ref=self.ref_reads if self.ref_reads != None else ".", alt=self.alt_reads if self.alt_reads != None else "."))


    def get_vcf_entry_as_dup(self, read_names = False, zmws = False):
        contig = self.source_contig
        start = self.source_start
        end = self.source_end
        length = self.source_end - self.source_start
        svtype = "DUP:TANDEM"
        if self.ref_reads != None and self.alt_reads != None:
            dp_string = str(self.ref_reads + self.alt_reads)
        else:
            dp_string = "."
        filters = []
        if self.genotype == "0/0":
            filters.append("hom_ref")
        if not(self.fully_covered):
            filters.append("not_fully_covered")
        info_template="SVTYPE={0};END={1};SVLEN={2};SUPPORT={3};STD_SPAN={4};STD_POS={5}"
        info_string = info_template.format(svtype, 
                                           end, 
                                           length, 
                                           len(set([sig.read for sig in self.members])), 
                                           self.get_std_span(), 
                                           self.get_std_pos())
        read_ids = [member.read for member in self.members]
        if read_names:
            info_string += ";READS={0}".format(",".join(read_ids))
        if zmws:
            valid_pacbio_names = True
            zmw_list = set()
            for read_id in read_ids:
                fields = read_id.split("/")
                if len(fields) != 3:
                    valid_pacbio_names = False
                    break
                zmw_list.add("/".join(fields[0:2]))
            if valid_pacbio_names:
                info_string += ";ZMWS={0}".format(len(zmw_list))
        return "{chrom}\t{pos}\t{id}\t{ref}\t{alt}\t{qual}\t{filter}\t{info}\t{format}\t{samples}".format(
                    chrom=contig,
                    pos=start+1,
                    id="PLACEHOLDERFORID",
                    ref="N",
                    alt="<" + svtype + ">",
                    qual=int(self.score),
                    filter="PASS" if len(filters) == 0 else ";".join(filters),
                    info=info_string,
                    format="GT:CN:DP:AD",
                    samples="{gt}:{cn}:{dp}:{ref},{alt}".format(gt=self.genotype, cn=self.copies + 1, dp=dp_string, ref=self.ref_reads if self.ref_reads != None else ".", alt=self.alt_reads if self.alt_reads != None else "."))


class CandidateDuplicationInterspersed(Candidate):
    def __init__(self, source_contig, source_start, source_end, dest_contig, dest_start, dest_end, members, score, std_span, std_pos, cutpaste=False, support_fraction = ".", genotype = "./.", ref_reads = None, alt_reads = None):
        self.source_contig = source_contig
        #0-based start of the region (first copied base)
        self.source_start = max(0, source_start)
        #0-based end of the region (one past the last copied base)
        self.source_end = source_end

        self.dest_contig = dest_contig
        #0-based start of the insertion (base after the insertion)
        self.dest_start = max(0, dest_start)
        #0-based end of the insertion (base after the insertion) + length of the insertion
        self.dest_end = dest_end

        self.members = members
        self.score = score
        self.std_span = std_span
        self.std_pos = std_pos
        self.cutpaste= cutpaste
        self.type = "DUP_INT"
        self.support_fraction = support_fraction
        self.genotype = genotype
        self.ref_reads = ref_reads
        self.alt_reads = alt_reads


    def get_destination(self):
        return (self.dest_contig, self.dest_start, self.dest_end)


    def get_bed_entries(self, sep="\t"):
        source_contig, source_start, source_end = self.get_source()
        dest_contig, dest_start, dest_end = self.get_destination()
        source_entry = sep.join(["{0}", "{1}", "{2}", "{3}", "{4}", "{5}", "{6}"]).format(source_contig, source_start,
                                                                                     source_end,
                                                                                     "int_dup_source;>{0}:{1}-{2};{3};{4}".format(
                                                                                         dest_contig, dest_start,
                                                                                         dest_end, self.get_std_span(), self.get_std_pos()), self.score, "origin potentially deleted" if self.cutpaste else ".",
                                                                                     "[" + "][".join(
                                                                                         [ev.as_string("|") for ev in
                                                                                          self.members]) + "]")
        dest_entry = sep.join(["{0}", "{1}", "{2}", "{3}", "{4}", "{5}", "{6}"]).format(dest_contig, dest_start, dest_end,
                                                                                   "int_dup_dest;<{0}:{1}-{2};{3};{4}".format(
                                                                                       source_contig, source_start,
                                                                                       source_end, self.get_std_span(), self.get_std_pos()), self.score, "origin potentially deleted" if self.cutpaste else ".",
                                                                                   "[" + "][".join(
                                                                                       [ev.as_string("|") for ev in
                                                                                        self.members]) + "]")
        return (source_entry, dest_entry)


    def get_vcf_entry_as_ins(self, sequence_alleles = False, reference = None, read_names = False, zmws = False):
        source_contig, source_start, source_end = self.get_source()
        dest_contig, dest_start, dest_end = self.get_destination()
        svtype = "INS"
        if self.ref_reads != None and self.alt_reads != None:
            dp_string = str(self.ref_reads + self.alt_reads)
        else:
            dp_string = "."
        filters = []
        if self.genotype == "0/0":
            filters.append("hom_ref")
        if sequence_alleles:
            ref_allele = reference.fetch(dest_contig, max(0, dest_start-1), max(0, dest_start-1) + 1).upper()
            alt_allele = ref_allele + reference.fetch(source_contig, source_start, source_end).upper()
        else:
            ref_allele = "N"
            alt_allele = "<" + self.type + ">"
        info_template="SVTYPE={0};{1}END={2};SVLEN={3};SUPPORT={4};STD_SPAN={5};STD_POS={6}"
        info_string = info_template.format(svtype,
                                           "CUTPASTE;" if self.cutpaste else "",
                                           dest_start,
                                           dest_end - dest_start,
                                           len(set([sig.read for sig in self.members])),
                                           self.get_std_span(),
                                           self.get_std_pos())
        read_ids = [member.read for member in self.members]
        if read_names:
            info_string += ";READS={0}".format(",".join(read_ids))
        if zmws:
            valid_pacbio_names = True
            zmw_list = set()
            for read_id in read_ids:
                fields = read_id.split("/")
                if len(fields) != 3:
                    valid_pacbio_names = False
                    break
                zmw_list.add("/".join(fields[0:2]))
            if valid_pacbio_names:
                info_string += ";ZMWS={0}".format(len(zmw_list))
        return "{chrom}\t{pos}\t{id}\t{ref}\t{alt}\t{qual}\t{filter}\t{info}\t{format}\t{samples}".format(
                    chrom=dest_contig,
                    pos=max(1, dest_start),
                    id="PLACEHOLDERFORID",
                    ref=ref_allele,
                    alt=alt_allele,
                    qual=int(self.score),
                    filter="PASS" if len(filters) == 0 else ";".join(filters),
                    info=info_string,
                    format="GT:DP:AD",
                    samples="{gt}:{dp}:{ref},{alt}".format(gt=self.genotype, dp=dp_string, ref=self.ref_reads if self.ref_reads != None else ".", alt=self.alt_reads if self.alt_reads != None else "."))


    def get_vcf_entry_as_dup(self, read_names = False, zmws = False):
        contig, start, end = self.get_source()
        svtype = "DUP:INT"
        if self.ref_reads != None and self.alt_reads != None:
            dp_string = str(self.ref_reads + self.alt_reads)
        else:
            dp_string = "."
        filters = []
        if self.genotype == "0/0":
            filters.append("hom_ref")
        info_template="SVTYPE={0};{1}END={2};SVLEN={3};SUPPORT={4};STD_SPAN={5};STD_POS={6}"
        info_string = info_template.format(svtype, 
                                           "CUTPASTE;" if self.cutpaste else "", 
                                           end, 
                                           end - start, 
                                           len(set([sig.read for sig in self.members])), 
                                           self.get_std_span(), 
                                           self.get_std_pos())
        read_ids = [member.read for member in self.members]
        if read_names:
            info_string += ";READS={0}".format(",".join(read_ids))
        if zmws:
            valid_pacbio_names = True
            zmw_list = set()
            for read_id in read_ids:
                fields = read_id.split("/")
                if len(fields) != 3:
                    valid_pacbio_names = False
                    break
                zmw_list.add("/".join(fields[0:2]))
            if valid_pacbio_names:
                info_string += ";ZMWS={0}".format(len(zmw_list))
        return "{chrom}\t{pos}\t{id}\t{ref}\t{alt}\t{qual}\t{filter}\t{info}\t{format}\t{samples}".format(
                    chrom=contig,
                    pos=start+1,
                    id="PLACEHOLDERFORID",
                    ref="N",
                    alt="<" + svtype + ">",
                    qual=int(self.score),
                    filter="PASS" if len(filters) == 0 else ";".join(filters),
                    info=info_string,
                    format="GT:DP:AD",
                    samples="{gt}:{dp}:{ref},{alt}".format(gt=self.genotype, dp=dp_string, ref=self.ref_reads if self.ref_reads != None else ".", alt=self.alt_reads if self.alt_reads != None else "."))


class CandidateBreakend(Candidate):
    def __init__(self, source_contig, source_start, source_direction, dest_contig, dest_start, dest_direction, members, score, std_pos1, std_pos2, support_fraction = ".", genotype = "./.", ref_reads = None, alt_reads = None):
        self.source_contig = source_contig
        #0-based source of the translocation (first base before the translocation)
        self.source_start = max(0, source_start)
        self.source_direction = source_direction

        self.dest_contig = dest_contig
        #0-based destination of the translocation (first base after the translocation)
        self.dest_start = max(0, dest_start)
        self.dest_direction = dest_direction

        self.members = members
        self.score = score
        self.std_pos1 = std_pos1
        self.std_pos2 = std_pos2
        self.type = "BND"
        self.support_fraction = support_fraction
        self.genotype = genotype
        self.ref_reads = ref_reads
        self.alt_reads = alt_reads


    def get_source(self):
        return (self.source_contig, self.source_start)


    def get_destination(self):
        return (self.dest_contig, self.dest_start)


    def get_std_pos1(self, ndigits=2):
        if self.std_pos1:
            return round(self.std_pos1, ndigits)
        else:
            return "."


    def get_std_pos2(self, ndigits=2):
        if self.std_pos2:
            return round(self.std_pos2, ndigits)
        else:
            return "."


    def get_bed_entries(self, sep="\t"):
        source_contig, source_start = self.get_source()
        dest_contig, dest_start = self.get_destination()
        source_entry = sep.join(["{0}", "{1}", "{2}", "{3}", "{4}", "{5}"]).format(source_contig, 
                                                                                   source_start,
                                                                                   source_start + 1,
                                                                                   "bnd;>{0}:{1};{2};{3}".format(
                                                                                        dest_contig, dest_start, self.get_std_pos1(), self.get_std_pos2()), 
                                                                                   self.score, 
                                                                                   "[" + "][".join(
                                                                                        [ev.as_string("|") for ev in self.members]) + "]")
        dest_entry = sep.join(["{0}", "{1}", "{2}", "{3}", "{4}", "{5}"]).format(dest_contig, 
                                                                                 dest_start, 
                                                                                 dest_start + 1,
                                                                                 "bnd;<{0}:{1};{2};{3}".format(
                                                                                        source_contig, source_start, self.get_std_pos1(), self.get_std_pos2()), 
                                                                                 self.score,
                                                                                 "[" + "][".join(
                                                                                        [ev.as_string("|") for ev in self.members]) + "]")
        return (source_entry, dest_entry)


    def get_vcf_entry(self, read_names = False, zmws = False):
        source_contig, source_start = self.get_source()
        dest_contig, dest_start = self.get_destination()
        if (self.source_direction == 'fwd') and (self.dest_direction == 'fwd'):
            alt_string = "N[{contig}:{start}[".format(contig = dest_contig, start = dest_start+1)
        elif (self.source_direction == 'fwd') and (self.dest_direction == 'rev'):
            alt_string = "N]{contig}:{start}]".format(contig = dest_contig, start = dest_start+1)
        elif (self.source_direction == 'rev') and (self.dest_direction == 'rev'):
            alt_string = "]{contig}:{start}]N".format(contig = dest_contig, start = dest_start+1)
        elif (self.source_direction == 'rev') and (self.dest_direction == 'fwd'):
            alt_string = "[{contig}:{start}[N".format(contig = dest_contig, start = dest_start+1)
        if self.ref_reads != None and self.alt_reads != None:
            dp_string = str(self.ref_reads + self.alt_reads)
        else:
            dp_string = "."
        filters = []
        if self.genotype == "0/0":
            filters.append("hom_ref")
        info_template="SVTYPE={0};SUPPORT={1};STD_POS1={2};STD_POS2={3}"
        info_string = info_template.format(self.type, 
                                           len(set([sig.read for sig in self.members])), 
                                           self.get_std_pos1(), 
                                           self.get_std_pos2())
        read_ids = [member.read for member in self.members]
        if read_names:
            info_string += ";READS={0}".format(",".join(read_ids))
        if zmws:
            valid_pacbio_names = True
            zmw_list = set()
            for read_id in read_ids:
                fields = read_id.split("/")
                if len(fields) != 3:
                    valid_pacbio_names = False
                    break
                zmw_list.add("/".join(fields[0:2]))
            if valid_pacbio_names:
                info_string += ";ZMWS={0}".format(len(zmw_list))
        return "{chrom}\t{pos}\t{id}\t{ref}\t{alt}\t{qual}\t{filter}\t{info}\t{format}\t{samples}".format(
                    chrom=source_contig,
                    pos=source_start+1,
                    id="PLACEHOLDERFORID",
                    ref="N",
                    alt=alt_string,
                    qual=int(self.score),
                    filter="PASS" if len(filters) == 0 else ";".join(filters),
                    info=info_string,
                    format="GT:DP:AD",
                    samples="{gt}:{dp}:{ref},{alt}".format(gt=self.genotype, dp=dp_string, ref=self.ref_reads if self.ref_reads != None else ".", alt=self.alt_reads if self.alt_reads != None else "."))


    def get_vcf_entry_reverse(self, read_names = False, zmws = False):
        source_contig, source_start = self.get_destination()
        dest_contig, dest_start = self.get_source()
        if (self.source_direction == 'rev') and (self.dest_direction == 'rev'):
            alt_string = "N[{contig}:{start}[".format(contig = dest_contig, start = dest_start+1)
        elif (self.source_direction == 'fwd') and (self.dest_direction == 'rev'):
            alt_string = "N]{contig}:{start}]".format(contig = dest_contig, start = dest_start+1)
        elif (self.source_direction == 'fwd') and (self.dest_direction == 'fwd'):
            alt_string = "]{contig}:{start}]N".format(contig = dest_contig, start = dest_start+1)
        elif (self.source_direction == 'rev') and (self.dest_direction == 'fwd'):
            alt_string = "[{contig}:{start}[N".format(contig = dest_contig, start = dest_start+1)
        if self.ref_reads != None and self.alt_reads != None:
            dp_string = str(self.ref_reads + self.alt_reads)
        else:
            dp_string = "."
        filters = []
        if self.genotype == "0/0":
            filters.append("hom_ref")
        info_template="SVTYPE={0};SUPPORT={1};STD_POS1={2};STD_POS2={3}"
        info_string = info_template.format(self.type, 
                                           len(set([sig.read for sig in self.members])), 
                                           self.get_std_pos2(), 
                                           self.get_std_pos1())
        read_ids = [member.read for member in self.members]
        if read_names:
            info_string += ";READS={0}".format(",".join(read_ids))
        if zmws:
            valid_pacbio_names = True
            zmw_list = set()
            for read_id in read_ids:
                fields = read_id.split("/")
                if len(fields) != 3:
                    valid_pacbio_names = False
                    break
                zmw_list.add("/".join(fields[0:2]))
            if valid_pacbio_names:
                info_string += ";ZMWS={0}".format(len(zmw_list))
        return "{chrom}\t{pos}\t{id}\t{ref}\t{alt}\t{qual}\t{filter}\t{info}\t{format}\t{samples}".format(
                    chrom=source_contig,
                    pos=source_start+1,
                    id="PLACEHOLDERFORID",
                    ref="N",
                    alt=alt_string,
                    qual=int(self.score),
                    filter="PASS" if len(filters) == 0 else ";".join(filters),
                    info=info_string,
                    format="GT:DP:AD",
                    samples="{gt}:{dp}:{ref},{alt}".format(gt=self.genotype, dp=dp_string, ref=self.ref_reads if self.ref_reads != None else ".", alt=self.alt_reads if self.alt_reads != None else "."))



import unittest
import tempfile

from svim.SVIM_input_parsing import guess_file_type, parse_arguments

class TestInputParsing(unittest.TestCase):

    def test_guess_file_type(self):
        fasta_paths = ["/test/path/test.file.fa", "/test/path/test.file.fasta", "/test/path/test.file.FA"]
        fastq_paths = ["/test/path/test.file.fq", "/test/path/test.file.FQ", "/test/path/test.file.fastq"]

        for p in fasta_paths:
            self.assertEqual(guess_file_type(p), "fasta")
            zipped_path1 = p + ".gz"
            zipped_path2 = p + ".gzip"
            self.assertEqual(guess_file_type(zipped_path1), "fasta_gzip")
            self.assertEqual(guess_file_type(zipped_path2), "fasta_gzip")
            list_path = p + ".fn"
            self.assertEqual(guess_file_type(list_path), "list")
        
        for p in fastq_paths:
            self.assertEqual(guess_file_type(p), "fastq")
            zipped_path1 = p + ".gz"
            zipped_path2 = p + ".gzip"
            self.assertEqual(guess_file_type(zipped_path1), "fastq_gzip")
            self.assertEqual(guess_file_type(zipped_path2), "fastq_gzip")
            list_path = p + ".fn"
            self.assertEqual(guess_file_type(list_path), "list")



import unittest

from svim.SVSignature import SignatureDeletion,SignatureInsertion

class TestSVSignature(unittest.TestCase):

    def test_accessors(self):
        deletion = SignatureDeletion("chr1", 100, 300, "cigar", "read1")

        self.assertEqual(deletion.get_source(), ("chr1", 100, 300))
        self.assertEqual(deletion.get_key(), ("DEL", "chr1", 300))
    
    def test_downstream_distance_to(self):
        deletion1 = SignatureDeletion("chr1", 100, 300, "cigar", "read1")
        deletion2 = SignatureDeletion("chr1", 450, 500, "cigar", "read2")
        deletion3 = SignatureDeletion("chr1", 150, 200, "cigar", "read3")
        deletion4 = SignatureDeletion("chr2", 350, 400, "cigar", "read3")
        insertion = SignatureInsertion("chr1", 150, 200, "cigar", "read2", "ACGTAGTAGCTAGCTTTGCTAGCATTAGCGACTGCTTACGCAGCTCCCTA")

        self.assertEqual(deletion1.downstream_distance_to(deletion2), 150)
        self.assertEqual(deletion1.downstream_distance_to(deletion3), 0)
        self.assertEqual(deletion1.downstream_distance_to(deletion4), float("Inf"))
        self.assertEqual(deletion1.downstream_distance_to(insertion), float("Inf"))
    
    def test_as_string(self):
        deletion1 = SignatureDeletion("chr1", 100, 300, "cigar", "read1")

        self.assertEqual(deletion1.as_string(), "chr1\t100\t300\tDEL;cigar\tread1")
        self.assertEqual(deletion1.as_string(":"), "chr1:100:300:DEL;cigar:read1")


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import Mock
from random import choices

from svim.SVIM_COMBINE import generate_insertion_consensus
from svim.SVSignature import SignatureInsertion, SignatureClusterUniLocal

class TestSVIMConsensus(unittest.TestCase):
    def setUp(self):
        self.nucleotides = ["A", "C", "G", "T"]
        #Prepare reference
        self.genome = "A"*100 + "C"*100
        self.reference = Mock()        
        self.reference.fetch = lambda contig, start, end: self.genome[start:end]
        
    def test_skipping(self):
        insertion_signatures = []
        for i in range(10):
            insertion_signatures.append(SignatureInsertion("chr1", 100, 100100, "suppl", "read"+str(i), "".join(choices(self.nucleotides, k=100000))))
        cluster_long = SignatureClusterUniLocal("chr1", 100, 100100, 10, 10, insertion_signatures, "INS", 0, 0)
        status_code, _ = generate_insertion_consensus(cluster_long, self.reference, maximum_haplotype_length = 10000)
        self.assertEqual(status_code, 1)
    
    def test_simple(self):
        #Prepare cluster
        insertion_sequence = "".join(choices(self.nucleotides, k=100))
        insertion_signatures = []
        for i in range(10):
            insertion_signatures.append(SignatureInsertion("chr1", 100, 200, "cigar", "read"+str(i), insertion_sequence))
        cluster_simple = SignatureClusterUniLocal("chr1", 100, 200, 10, 10, insertion_signatures, "INS", 0, 0)
        status_code, result = generate_insertion_consensus(cluster_simple, self.reference)
        self.assertEqual(status_code, 0)
        realigned_insertion_start, realigned_insertion_size, insertion_consensus = result
        self.assertEqual(realigned_insertion_start, 100)
        self.assertEqual(realigned_insertion_size, 100)
        self.assertEqual(insertion_consensus, insertion_sequence)

if __name__ == '__main__':
    unittest.main()



from __future__ import print_function

import sys

from svim.SVSignature import SignatureDeletion, SignatureInsertion, SignatureTranslocation


def analyze_cigar_indel(tuples, min_length):
    """Parses CIGAR tuples (op, len) and returns Indels with a length > minLength"""
    pos_ref = 0
    pos_read = 0
    indels = []
    for operation, length in tuples:
        if operation == 0:                     # alignment match
            pos_ref += length
            pos_read += length
        elif operation == 1:                   # insertion
            if length >= min_length:
                indels.append((pos_ref, pos_read, length, "INS"))
            pos_read += length
        elif operation == 2:                   # deletion
            if length >= min_length:
                indels.append((pos_ref, pos_read, length, "DEL"))
            pos_ref += length
        elif operation == 4:                   # soft clip
            pos_read += length
        elif operation == 7 or operation == 8:        # match or mismatch
            pos_ref += length
            pos_read += length
    return indels


def analyze_alignment_indel(alignment, bam, query_name, options):
    sv_signatures = []
    #Translocation signatures from other SV classes are stored separately for --all_bnd option
    translocation_signatures_all_bnds = []
    ref_chr = bam.getrname(alignment.reference_id)
    ref_start = alignment.reference_start
    indels = analyze_cigar_indel(alignment.cigartuples, options.min_sv_size)
    for pos_ref, pos_read, length, typ in indels:
        if typ == "DEL":
            sv_signatures.append(SignatureDeletion(ref_chr, ref_start + pos_ref, ref_start + pos_ref + length, "cigar", query_name))
            if options.all_bnds:
                translocation_signatures_all_bnds.append(SignatureTranslocation(ref_chr, ref_start + pos_ref, 'fwd', ref_chr, ref_start + pos_ref + length, 'fwd', "cigar", query_name))
        elif typ == "INS":
            try:
                insertion_seq = alignment.query_sequence[pos_read:pos_read+length]
            except TypeError:
                insertion_seq = ""
            sv_signatures.append(SignatureInsertion(ref_chr, ref_start + pos_ref, ref_start + pos_ref + length, "cigar", query_name, insertion_seq))
    return sv_signatures, translocation_signatures_all_bnds





import unittest
import tempfile

from random import uniform, choices

from svim.SVIM_clustering import form_partitions, partition_and_cluster
from svim.SVSignature import SignatureDeletion
from svim.SVIM_input_parsing import parse_arguments

class TestSVIMClustering(unittest.TestCase):
    def setUp(self):
        self.signatures = []
        #group 0
        for i in range(10):
            center = 100000 + uniform(-100, 100)
            half_span = 1000 + uniform(-100, 100)
            new_sig = SignatureDeletion("chr1", center - half_span, center + half_span, "cigar", str(i))
            self.signatures.append(new_sig)
        #group 1
        for i in range(10, 20):
            center = 200000 + uniform(-100, 100)
            half_span = 1000 + uniform(-100, 100)
            new_sig = SignatureDeletion("chr1", center - half_span, center + half_span, "cigar", str(i))
            self.signatures.append(new_sig)
        #group 2
        for i in range(20, 30):
            center = 100000 + uniform(-100, 100)
            half_span = 2000 + uniform(-100, 100)
            new_sig = SignatureDeletion("chr1", center - half_span, center + half_span, "cigar", str(i))
            self.signatures.append(new_sig)
        self.temp_genome = tempfile.NamedTemporaryFile(mode="w")
        self.temp_genome.write('>chr1\n')
        self.temp_genome.write("".join(choices(["A", "C", "G", "T"], k=300000)))
        self.options = parse_arguments('1.5.0', ['alignment', 'myworkdir', 'mybamfile', self.temp_genome.name])

    def tearDown(self):
        self.temp_genome.close()

    def test_partitioning(self):
        partitions = form_partitions(self.signatures, 100)
        self.assertEqual(len(partitions), 2)
        for partition in partitions:
            groups_in_partition = set([int(member.read) // 10 for member in partition])
            self.assertTrue(groups_in_partition in [set([0,2]), set([1])])

    def test_partitioning_large_distance(self):
        partitions = form_partitions(self.signatures, 100000)
        self.assertEqual(len(partitions), 1)
        groups_in_partition = set([int(member.read) // 10 for member in partitions[0]])
        self.assertEqual(groups_in_partition, set([0,1,2]))

    def test_clustering(self):
        clusters = partition_and_cluster(self.signatures, options=self.options, type="deleted regions")
        self.assertEqual(len(clusters), 3)
        for cluster in clusters:
            self.assertEqual(len(set([int(member.read) // 10 for member in cluster.members])), 1)
    
    def test_scores(self):
        clusters = partition_and_cluster(self.signatures, options=self.options, type="deleted regions")
        for cluster in clusters:
            self.assertGreaterEqual(cluster.score, 10)
            self.assertLessEqual(cluster.score, 10 + 20/8)

if __name__ == '__main__':
    unittest.main()



from __future__ import print_function

import sys
import logging

from random import seed, sample
from statistics import mean, stdev
import numpy as np
from scipy.cluster.hierarchy import linkage, fcluster
from edlib import align
from pysam import FastaFile

from svim.SVSignature import SignatureClusterUniLocal, SignatureClusterBiLocal
from svim.SVCandidate import CandidateDuplicationInterspersed


def form_partitions(sv_signatures, max_distance):
    """Form partitions of signatures using mean distance."""
    sorted_signatures = sorted(sv_signatures, key=lambda evi: evi.get_key())
    partitions = []
    current_partition = []
    for signature in sorted_signatures:
        if len(current_partition) > 0 and current_partition[-1].downstream_distance_to(signature) > max_distance:
            partitions.append(current_partition[:])
            current_partition = []
        current_partition.append(signature)
    if len(current_partition) > 0:
        partitions.append(current_partition[:])
    return partitions


def compute_haplotype_edit_distance(signature1, signature2, reference, window_padding = 100):
    window_start = min(signature1.start, signature2.start) - window_padding
    window_end = max(signature1.start, signature2.start) + window_padding

    #construct haplotype sequences for both signatures
    haplotype1 = reference.fetch(signature1.contig, max(0, window_start), max(0, signature1.start)).upper()
    haplotype1 += signature1.sequence.upper()
    haplotype1 += reference.fetch(signature1.contig, max(0, signature1.start), max(0, window_end)).upper()

    haplotype2 = reference.fetch(signature2.contig, max(0, window_start), max(0, signature2.start)).upper()
    haplotype2 += signature2.sequence.upper()
    haplotype2 += reference.fetch(signature2.contig, max(0, signature2.start), max(0, window_end)).upper()

    return align(haplotype1, haplotype2)["editDistance"]

def span_position_distance(signature1, signature2, signature_type, reference, position_distance_normalizer, edit_distance_normalizer, cluster_max_distance):
    if signature_type == "DEL" or signature_type == "DUP_TAN":
        span1 = signature1.get_source()[2] - signature1.get_source()[1]
        span2 = signature2.get_source()[2] - signature2.get_source()[1]
        center1 = (signature1.get_source()[1] + signature1.get_source()[2]) // 2
        center2 = (signature2.get_source()[1] + signature2.get_source()[2]) // 2
        position_distance = abs(center1 - center2) / position_distance_normalizer
        span_distance = abs(span1 - span2) / max(span1, span2)
        return position_distance + span_distance
    elif signature_type == "INV": #two signatures from same read can be clustered together"
        span1 = signature1.get_source()[2] - signature1.get_source()[1]
        span2 = signature2.get_source()[2] - signature2.get_source()[1]
        center1 = (signature1.get_source()[1] + signature1.get_source()[2]) // 2
        center2 = (signature2.get_source()[1] + signature2.get_source()[2]) // 2
        position_distance = abs(center1 - center2) / position_distance_normalizer
        span_distance = abs(span1 - span2) / max(span1, span2)
        return position_distance + span_distance
    elif signature_type == "INS": #center is the insertion location
        span1 = signature1.get_source()[2] - signature1.get_source()[1]
        span2 = signature2.get_source()[2] - signature2.get_source()[1]
        center1 = signature1.get_source()[1]
        center2 = signature2.get_source()[1]
        position_distance = abs(center1 - center2) / position_distance_normalizer
        if position_distance > 2*cluster_max_distance:
            #do not compute edit distance if insertions are too distant
            span_distance = abs(span1 - span2) / max(span1, span2)
            return position_distance + span_distance
        else:
            edit_distance = compute_haplotype_edit_distance(signature1, signature2, reference)
            sequence_distance = edit_distance / max(span1, span2) / edit_distance_normalizer
            return position_distance + sequence_distance
    elif signature_type == "DUP_INT": #position distance is computed for source and destination
        span1 = signature1.get_source()[2] - signature1.get_source()[1]
        span2 = signature2.get_source()[2] - signature2.get_source()[1]
        source_center1 = (signature1.get_source()[1] + signature1.get_source()[2]) // 2
        source_center2 = (signature2.get_source()[1] + signature2.get_source()[2]) // 2
        position_distance_source = abs(source_center1 - source_center2) / position_distance_normalizer
        position_distance_destination = abs(signature1.get_destination()[1] - signature2.get_destination()[1]) / position_distance_normalizer
        span_distance = abs(span1 - span2) / max(span1, span2)
        return position_distance_source + position_distance_destination + span_distance
    elif signature_type == "BND": #only position distance is computed
        dist1 = abs(signature1.get_source()[1] - signature2.get_source()[1])
        dist2 = abs(signature1.get_destination()[1] - signature2.get_destination()[1])
        if signature1.direction1 == signature2.direction1 and signature1.direction2 == signature2.direction2:
            position_distance = (dist1 + dist2) / 3000
        else:
            position_distance = 99999
        return position_distance
    else:
        return None


def span_position_distance_clusters(cluster1, cluster2, position_distance_normalizer):
    "Span position distance function for merging clusters"
    span1 = cluster1.get_source()[2] - cluster1.get_source()[1]
    span2 = cluster2.get_source()[2] - cluster2.get_source()[1]
    center1 = (cluster1.get_source()[1] + cluster1.get_source()[2]) // 2
    center2 = (cluster2.get_source()[1] + cluster2.get_source()[2]) // 2
    position_distance = abs(center1 - center2) / position_distance_normalizer
    span_distance = abs(span1 - span2) / max(span1, span2)    
    return position_distance + span_distance


def span_position_distance_intdup_candidates(signature1, signature2, position_distance_normalizer):
    "Span position distance function for clustering candidates"
    span1 = signature1.get_source()[2] - signature1.get_source()[1]
    span2 = signature2.get_source()[2] - signature2.get_source()[1]
    source_center1 = (signature1.get_source()[1] + signature1.get_source()[2]) // 2
    source_center2 = (signature2.get_source()[1] + signature2.get_source()[2]) // 2
    position_distance_source = abs(source_center1 - source_center2) / position_distance_normalizer
    position_distance_destination = abs(signature1.get_destination()[1] - signature2.get_destination()[1]) / position_distance_normalizer
    span_distance = abs(span1 - span2) / max(span1, span2)
    return position_distance_source + position_distance_destination + span_distance


def clusters_from_partitions(partitions, reference, options):
    """Finds clusters in partitions using span-position distance and hierarchical clustering. 
    Assumes that all signatures in the given partition are of the same type and on the same contig"""
    clusters_final = []
    large_partitions = 0
    duplicate_signatures = 0
    #initialize random number generator with fixed number to produce same output from same input
    seed(1524)
    # Find clusters in each partition individually.
    for partition in partitions:
        if len(partition) > 100:
            partition_sample = sample(partition, 100)
            large_partitions += 1
        else:
            partition_sample = partition
        element_type = partition_sample[0].type
        assert(element_type in ["DEL", "DUP_TAN", "INV", "INS", "DUP_INT", "BND"])

        #remove similar signatures coming from the same read
        if element_type == "INV":
            #no duplication removal for inversions because they consist of two complementary signatures from the same read
            partition_sample_without_duplicates = partition_sample
        else:
            duplicates_from_same_read = set()
            for i in range(len(partition_sample)-1):
                for j in range(i+1, len(partition_sample)):
                    if partition_sample[i].read == partition_sample[j].read and span_position_distance(partition_sample[i], partition_sample[j], element_type, reference, options.position_distance_normalizer, options.edit_distance_normalizer, options.cluster_max_distance) <= options.cluster_max_distance:
                        duplicates_from_same_read.add(j)
            duplicate_signatures += len(duplicates_from_same_read)
            partition_sample_without_duplicates = [partition_sample[i] for i in range(len(partition_sample)) if i not in duplicates_from_same_read]

        if len(partition_sample_without_duplicates) == 1:
            clusters_final.append([partition_sample_without_duplicates[0]])
            continue

        #compute pairwise distances
        distances = []
        if element_type == "INV":
            for i in range(len(partition_sample_without_duplicates)-1):
                for j in range(i+1, len(partition_sample_without_duplicates)):
                    distances.append(span_position_distance(partition_sample_without_duplicates[i], partition_sample_without_duplicates[j], element_type, reference, options.position_distance_normalizer, options.edit_distance_normalizer, options.cluster_max_distance))
        else:
            for i in range(len(partition_sample_without_duplicates)-1):
                for j in range(i+1, len(partition_sample_without_duplicates)):
                    if partition_sample_without_duplicates[i].read == partition_sample_without_duplicates[j].read:
                        distances.append(99999)
                    else:
                        distances.append(span_position_distance(partition_sample_without_duplicates[i], partition_sample_without_duplicates[j], element_type, reference, options.position_distance_normalizer, options.edit_distance_normalizer, options.cluster_max_distance))
        Z = linkage(np.array(distances), method = "average")
        cluster_indices = list(fcluster(Z, options.cluster_max_distance, criterion='distance'))
        new_clusters = [[] for i in range(max(cluster_indices))]
        for signature_index, cluster_index in enumerate(cluster_indices):
            new_clusters[cluster_index-1].append(partition_sample_without_duplicates[signature_index])
        clusters_final.extend(new_clusters)
    if len(partitions) > 0:
        if len(partitions[0]) > 0:
            logging.debug("%d out of %d partitions for %s exceeded 100 elements." % (large_partitions, len(partitions), partitions[0][0].type))
            logging.debug("%d %s signatures were removed due to similarity to another signature from the same read." % (duplicate_signatures, partitions[0][0].type))
    return clusters_final


def calculate_score(cluster, std_span, std_pos, span, type):
    if std_span == None or std_pos == None:
        span_deviation_score = 0
        pos_deviation_score = 0
    else:
        span_deviation_score = 1 - min(1, std_span / span)
        pos_deviation_score = 1 - min(1, std_pos / span)

    if type == "INV":
        directions = [signature.direction for signature in cluster]
        direction_counts = [0, 0, 0, 0, 0]
        for direction in directions:
            if direction == "left_fwd":
                direction_counts[0] += 1
            if direction == "left_rev":
                direction_counts[1] += 1
            if direction == "right_fwd":
                direction_counts[2] += 1
            if direction == "right_rev":
                direction_counts[3] += 1
            if direction == "all":
                direction_counts[4] += 1
        left_signatures = direction_counts[0] + direction_counts[1]
        right_signatures = direction_counts[2] + direction_counts[3]
        valid_signatures = min(left_signatures, right_signatures) + direction_counts[4]
        num_signatures = min(80, valid_signatures)
    else:
        num_signatures = min(80, len(cluster))
    return num_signatures + span_deviation_score * (num_signatures / 8) + pos_deviation_score * (num_signatures / 8)


def consolidate_clusters_unilocal(clusters):
    """Consolidate clusters to a list of (type, contig, mean start, mean end, cluster size, members) tuples."""
    consolidated_clusters = []
    for cluster in clusters:
        average_start = sum([member.get_source()[1] for member in cluster]) / len(cluster)
        average_end = sum([member.get_source()[2] for member in cluster]) / len(cluster)
        if len(cluster) > 1:
            std_span = stdev([member.get_source()[2] - member.get_source()[1] for member in cluster])
            std_pos = stdev([(member.get_source()[2] + member.get_source()[1]) / 2 for member in cluster])
        else:
            std_span = None
            std_pos = None
        score = calculate_score(cluster, std_span, std_pos, average_end - average_start, cluster[0].type)
        consolidated_clusters.append(SignatureClusterUniLocal(cluster[0].get_source()[0], int(round(average_start)), int(round(average_end)), score, len(cluster), cluster, cluster[0].type, std_span, std_pos))
    return consolidated_clusters


def consolidate_clusters_bilocal(clusters):
    """Consolidate clusters to a list of (type, contig, mean start, mean end, cluster size, members) tuples."""
    consolidated_clusters = []
    for cluster in clusters:
        #Source
        source_average_start = sum([member.get_source()[1] for member in cluster]) / len(cluster)
        source_average_end = sum([member.get_source()[2] for member in cluster]) / len(cluster)
        if len(cluster) > 1:
            source_std_span = stdev([member.get_source()[2] - member.get_source()[1] for member in cluster])
            source_std_pos = stdev([(member.get_source()[2] + member.get_source()[1]) / 2 for member in cluster])
        else:
            source_std_span = None
            source_std_pos = None

        if cluster[0].type == "DUP_TAN":
            max_copies = max([member.copies for member in cluster])
            score = calculate_score(cluster, source_std_span, source_std_pos, source_average_end - source_average_start, cluster[0].type)
            consolidated_clusters.append(SignatureClusterBiLocal(cluster[0].get_source()[0],
                                                                int(round(source_average_start)),
                                                                int(round(source_average_end)),
                                                                cluster[0].get_source()[0],
                                                                int(round(source_average_end)),
                                                                int(round(source_average_end)) + max_copies *
                                                                (int(round(source_average_end)) -
                                                                 int(round(source_average_start))),
                                                                score, len(cluster), cluster, cluster[0].type, source_std_span, source_std_pos))
        elif cluster[0].type == "DUP_INT":
            #Destination
            destination_average_start = sum([member.get_destination()[1] for member in cluster]) / len(cluster)
            destination_average_end = sum([member.get_destination()[2] for member in cluster]) / len(cluster)
            if len(cluster) > 1:
                destination_std_span = stdev([member.get_destination()[2] - member.get_destination()[1] for member in cluster])
                destination_std_pos = stdev([(member.get_destination()[2] + member.get_destination()[1]) / 2 for member in cluster])
            else:
                destination_std_span = None
                destination_std_pos = None
            if source_std_span == None or source_std_pos == None or destination_std_span == None or destination_std_pos == None:
                score = calculate_score(cluster, None, None, mean([source_average_end - source_average_start, destination_average_end - destination_average_start]), cluster[0].type)
                consolidated_clusters.append(SignatureClusterBiLocal(cluster[0].get_source()[0], int(round(source_average_start)), int(round(source_average_end)),
                                                                cluster[0].get_destination()[0], int(round(destination_average_start)), int(round(destination_average_end)), score, len(cluster), cluster, cluster[0].type, None, None))
            else:
                score = calculate_score(cluster, mean([source_std_span, destination_std_span]), mean([source_std_pos, destination_std_pos]), mean([source_average_end - source_average_start, destination_average_end - destination_average_start]), cluster[0].type)
                consolidated_clusters.append(SignatureClusterBiLocal(cluster[0].get_source()[0], int(round(source_average_start)), int(round(source_average_end)),
                                                                cluster[0].get_destination()[0], int(round(destination_average_start)), int(round(destination_average_end)), score, len(cluster), cluster, cluster[0].type, mean([source_std_span, destination_std_span]), mean([source_std_pos, destination_std_pos])))
        elif cluster[0].type == "BND":
            #Destination
            destination_average_start = sum([member.get_destination()[1] for member in cluster]) / len(cluster)
            destination_average_end = sum([member.get_destination()[2] for member in cluster]) / len(cluster)
            #Directions
            directions1 = list(set([member.direction1 for member in cluster]))
            assert len(directions1) == 1
            direction1 = directions1[0]
            directions2 = list(set([member.direction2 for member in cluster]))
            assert len(directions2) == 1
            direction2 = directions2[0]

            if len(cluster) > 1:
                destination_std_pos = stdev([(member.get_destination()[2] + member.get_destination()[1]) / 2 for member in cluster])
            else:
                destination_std_span = None
                destination_std_pos = None
            if source_std_pos == None or destination_std_pos == None:
                score = calculate_score(cluster, None, None, 500, cluster[0].type)
                new_signature_cluster = SignatureClusterBiLocal(cluster[0].get_source()[0], int(round(source_average_start)), int(round(source_average_end)),
                                                                cluster[0].get_destination()[0], int(round(destination_average_start)), int(round(destination_average_end)), score, len(cluster), cluster, cluster[0].type, None, None)
            else:
                score = calculate_score(cluster, source_std_pos, destination_std_pos, 500, cluster[0].type)
                new_signature_cluster = SignatureClusterBiLocal(cluster[0].get_source()[0], int(round(source_average_start)), int(round(source_average_end)),
                                                                cluster[0].get_destination()[0], int(round(destination_average_start)), int(round(destination_average_end)), score, len(cluster), cluster, cluster[0].type, source_std_pos, destination_std_pos)
            new_signature_cluster.direction1 = direction1
            new_signature_cluster.direction2 = direction2
            consolidated_clusters.append(new_signature_cluster)
    return consolidated_clusters


def partition_and_cluster_candidates(candidates, options, type):
    partitions = form_partitions(candidates, options.partition_max_distance)
    clusters = []
    large_partitions = 0
    #initialize random number generator with fixed number to produce same output from same input
    seed(1524)
    # Find clusters in each partition individually.
    for partition in partitions:
        if len(partition) == 1:
            clusters.append([partition[0]])
            continue
        elif len(partition) > 100:
            partition_sample = sample(partition, 100)
            large_partitions += 1
        else:
            partition_sample = partition
        element_type = partition_sample[0].type
        distances = []
        for i in range(len(partition_sample)-1):
            for j in range(i+1, len(partition_sample)):
                distances.append(span_position_distance_intdup_candidates(partition_sample[i], partition_sample[j], options.position_distance_normalizer))
        Z = linkage(np.array(distances), method = "average")

        cluster_indices = list(fcluster(Z, options.cluster_max_distance, criterion='distance'))
        new_clusters = [[] for i in range(max(cluster_indices))]
        for signature_index, cluster_index in enumerate(cluster_indices):
            new_clusters[cluster_index-1].append(partition_sample[signature_index])
        clusters.extend(new_clusters)
    if len(partitions) > 0:
        if len(partitions[0]) > 0:
            logging.debug("%d out of %d partitions for %s exceeded 100 elements." % (large_partitions, len(partitions), partitions[0][0].type))
    logging.info("Clustered {0}: {1} partitions and {2} clusters".format(type, len(partitions), len(clusters)))

    final_candidates = []
    for cluster in clusters:
        combined_score = max([candidate.score for candidate in cluster])
        combined_members = [member for candidate in cluster for member in candidate.members]

        stds_span = [candidate.std_span for candidate in cluster if candidate.std_span != None]
        if len(stds_span) >= 1:
            combined_std_span = mean(stds_span)
        else:
            combined_std_span = None
        stds_pos = [candidate.std_pos for candidate in cluster if candidate.std_pos != None]
        if len(stds_pos) >= 1:
            combined_std_pos = mean(stds_pos)
        else:
            combined_std_pos = None

        #Source
        source_average_start = (sum([candidate.get_source()[1] for candidate in cluster]) / len(cluster))
        source_average_end = (sum([candidate.get_source()[2] for candidate in cluster]) / len(cluster))

        #Destination
        destination_average_start = (sum([candidate.get_destination()[1] for candidate in cluster]) / len(cluster))
        destination_average_end = (sum([candidate.get_destination()[2] for candidate in cluster]) / len(cluster))

        #Origin deleted?
        cutpaste = False
        for member in cluster:
            if member.cutpaste:
                cutpaste = True

        if cluster[0].type == "DUP_INT":
            final_candidates.append(CandidateDuplicationInterspersed(cluster[0].get_source()[0], int(round(source_average_start)), int(round(source_average_end)),
                                                       cluster[0].get_destination()[0], int(round(destination_average_start)), int(round(destination_average_end)), combined_members, combined_score, combined_std_span, combined_std_pos, cutpaste))
    return final_candidates


def partition_and_cluster(signatures, options, type):
    partitions = form_partitions(signatures, options.partition_max_distance)
    with FastaFile(options.genome) as reference:
        clusters = clusters_from_partitions(partitions, reference, options)
    logging.info("Clustered {0}: {1} partitions and {2} clusters".format(type, len(partitions), len(clusters)))
    if type == "deleted regions" or type == "inserted regions" or type == "inverted regions":
        return sorted(consolidate_clusters_unilocal(clusters), key=lambda cluster: (cluster.contig, (cluster.end + cluster.start) / 2))
    elif type == "tandem duplicated regions" or type == "inserted regions with detected region of origin" or type == "translocation breakpoints":
        return consolidate_clusters_bilocal(clusters)
    else:
        logging.error("Unknown parameter type={0} to function partition_and_cluster.")



import unittest
import unittest.mock

from svim.SVCandidate import CandidateDeletion, CandidateInversion
from svim.SVSignature import SignatureDeletion

class DeletionCandidateTestCase(unittest.TestCase):
    def setUp(self):
        self.contig = "chr1"
        self.start = 1000
        self.end = 2000
        self.members = [SignatureDeletion(self.contig, self.start, self.end, "cigar", "read1")]
        self.score = 2
        self.std_span = 10.2346
        self.std_pos = 21.3453
        self.deletion = CandidateDeletion(self.contig,
                                          self.start,
                                          self.end,
                                          self.members,
                                          self.score,
                                          self.std_span,
                                          self.std_pos)
        self.distance = 1000
        self.size = 1000
        self.other_contig = 'chr9'
        self.deletion2 = CandidateDeletion(self.contig,
                                           self.end+self.distance,
                                           self.end+self.distance+self.size,
                                           self.members,
                                           self.score,
                                           self.std_span,
                                           self.std_pos)
        self.inversion = CandidateInversion(self.contig,
                                            self.end+self.distance,
                                            self.end+self.distance+self.size,
                                            self.members,
                                            self.score,
                                            self.std_span,
                                            self.std_pos)
        self.deletion3 = CandidateDeletion(self.other_contig,
                                           self.end+self.distance,
                                           self.end+self.distance+self.size,
                                           self.members,
                                           self.score,
                                           self.std_span,
                                           self.std_pos)

    def test_get_key(self):
        self.assertEqual(self.deletion.get_key(), ("DEL", self.contig, self.end),
                         'incorrect key')

    def test_get_source(self):
        self.assertEqual(self.deletion.get_source(), (self.contig, self.start, self.end),
                         'incorrect source')

    def test_get_std_span(self):
        self.assertEqual(self.deletion.get_std_span(), round(self.std_span, 2))
        self.assertEqual(self.deletion.get_std_span(3), round(self.std_span, 3))

    def test_get_std_pos(self):
        self.assertEqual(self.deletion.get_std_pos(), round(self.std_pos, 2))
        self.assertEqual(self.deletion.get_std_pos(3), round(self.std_pos, 3))

    def test_downstream_distance_to(self):
        self.assertEqual(self.deletion.downstream_distance_to(self.deletion2), self.distance,
                         'incorrect distance')
        self.assertEqual(self.deletion2.downstream_distance_to(self.deletion), 0,
                         'distance should be 0 for upstream variant')
        self.assertEqual(self.deletion.downstream_distance_to(self.inversion), float('Inf'),
                         'distance should be Inf for other variant types')
        self.assertEqual(self.deletion.downstream_distance_to(self.deletion3), float('Inf'),
                         'distance should be Inf for variant on other contig')

    def test_get_vcf_entry(self):
        vcf_string_normal = "\t".join(["chr1",
                                       "1000",
                                       "PLACEHOLDERFORID",
                                       "N",
                                       "<DEL>",
                                       "2",
                                       "PASS",
                                       "SVTYPE=DEL;END=2000;SVLEN=-1000;SUPPORT=1;STD_SPAN=10.23;STD_POS=21.35",
                                       "GT:DP:AD",
                                       "./.:.:.,."])
        self.assertEqual(self.deletion.get_vcf_entry(), vcf_string_normal,
                         'incorrect VCF record')

        ref_allele = "ACGTCGGATCGCAT"
        alt_allele = "A"
        reference = unittest.mock.Mock()
        reference.fetch.side_effect = [ref_allele, alt_allele]
        vcf_string_sequence_allele = "\t".join(["chr1",
                                                "1000",
                                                "PLACEHOLDERFORID",
                                                "ACGTCGGATCGCAT",
                                                "A",
                                                "2",
                                                "PASS",
                                                "SVTYPE=DEL;END=2000;SVLEN=-1000;SUPPORT=1;STD_SPAN=10.23;STD_POS=21.35",
                                                "GT:DP:AD",
                                                "./.:.:.,."])
        self.assertEqual(self.deletion.get_vcf_entry(sequence_alleles=True, reference=reference),
                         vcf_string_sequence_allele,
                         'incorrect VCF record with sequence alleles')



import matplotlib
import logging
import random
import math
matplotlib.use('Agg')
import matplotlib.pyplot as plt

def plot_sv_lengths(deletion_candidates, inversion_candidates, int_duplication_candidates, tan_dup_candidates, novel_insertion_candidates, options):
    len_dict_5 = dict()
    len_dict_10 = dict()
    len_dict_5["DEL"] = [v.get_source()[2] - v.get_source()[1] for v in deletion_candidates if v.score >= 5]
    len_dict_5["INV"] = [v.get_source()[2] - v.get_source()[1] for v in inversion_candidates if v.score >= 5]
    len_dict_5["DUP_INT"] = [v.get_destination()[2] - v.get_destination()[1] for v in int_duplication_candidates if v.score >= 5]
    len_dict_5["DUP_TAN"] = [v.get_destination()[2] - v.get_destination()[1] for v in tan_dup_candidates if v.score >= 5]
    len_dict_5["INS"] = [v.get_destination()[2] - v.get_destination()[1] for v in novel_insertion_candidates if v.score >= 5]
    draw_sv_length_plot(dict_of_lengths=len_dict_5, output=options.working_dir + "/sv-lengths-q5.png")
    len_dict_10["DEL"] = [v.get_source()[2] - v.get_source()[1] for v in deletion_candidates if v.score >= 10]
    len_dict_10["INV"] = [v.get_source()[2] - v.get_source()[1] for v in inversion_candidates if v.score >= 10]
    len_dict_10["DUP_INT"] = [v.get_destination()[2] - v.get_destination()[1] for v in int_duplication_candidates if v.score >= 10]
    len_dict_10["DUP_TAN"] = [v.get_destination()[2] - v.get_destination()[1] for v in tan_dup_candidates if v.score >= 10]
    len_dict_10["INS"] = [v.get_destination()[2] - v.get_destination()[1] for v in novel_insertion_candidates if v.score >= 10]
    draw_sv_length_plot(dict_of_lengths=len_dict_10, output=options.working_dir + "/sv-lengths-q10.png")


def draw_sv_length_plot(dict_of_lengths, output):
    """Makes two stacked bar charts
    Plotting two bar charts of number of SVs by length split by SV type
    Use a consistent colouring scheme for those in "standard_order" to
    make comparison reasonable

    First bar chart is up to 2kb with bins of 10bp
    Second bar chart is up to 20kb, with bins of 100bp
     and uses log scaling on the y-axis
    """
    standard_order = ['DEL', 'INS', 'INV', 'DUP_INT', 'DUP_TAN']
    names, lengths = zip(
            *sorted([(svtype, lengths) for svtype, lengths in dict_of_lengths.items()],
                    key=lambda x: standard_order.index(x[0])))
    plt.subplot(2, 1, 1)
    plt.hist(x=lengths,
             bins=[i for i in range(0, 2000, 10)],
             stacked=True,
             histtype='bar',
             label=names)
    plt.xlabel('Length of structural variant')
    plt.ylabel('Number of variants')
    plt.legend(frameon=False,
               fontsize="small")

    plt.subplot(2, 1, 2)
    plt.hist(x=lengths,
             bins=[i for i in range(0, 20000, 100)],
             stacked=True,
             histtype='bar',
             label=names,
             log=True)
    plt.xlabel('Length of structural variant')
    plt.ylabel('Number of variants')
    plt.legend(frameon=False,
               fontsize="small")
    plt.tight_layout()
    plt.savefig(output)
    plt.clf()


def plot_sv_alleles(candidates, options):
    refs_11 = [candidate.ref_reads for candidate in candidates if candidate.genotype == '1/1' and candidate.score >= 5 and candidate.ref_reads != None and candidate.alt_reads != None]
    alts_11 = [candidate.alt_reads for candidate in candidates if candidate.genotype == '1/1' and candidate.score >= 5 and candidate.ref_reads != None and candidate.alt_reads != None]
    refs_10 = [candidate.ref_reads for candidate in candidates if candidate.genotype == '0/1' and candidate.score >= 5 and candidate.ref_reads != None and candidate.alt_reads != None]
    alts_10 = [candidate.alt_reads for candidate in candidates if candidate.genotype == '0/1' and candidate.score >= 5 and candidate.ref_reads != None and candidate.alt_reads != None]
    refs_00 = [candidate.ref_reads for candidate in candidates if candidate.genotype == '0/0' and candidate.score >= 5 and candidate.ref_reads != None and candidate.alt_reads != None]
    alts_00 = [candidate.alt_reads for candidate in candidates if candidate.genotype == '0/0' and candidate.score >= 5 and candidate.ref_reads != None and candidate.alt_reads != None]
    refs_nn = [candidate.ref_reads for candidate in candidates if candidate.genotype == './.' and candidate.score >= 5 and candidate.ref_reads != None and candidate.alt_reads != None]
    alts_nn = [candidate.alt_reads for candidate in candidates if candidate.genotype == './.' and candidate.score >= 5 and candidate.ref_reads != None and candidate.alt_reads != None]

    draw_allele_plot(refs_11, alts_11, refs_10, alts_10, refs_00, alts_00, refs_nn, alts_nn, output=options.working_dir + "/sv-genotypes-q5.png")


def draw_allele_plot(refs_11, alts_11, refs_10, alts_10, refs_00, alts_00, refs_nn, alts_nn, output):
    """Makes a scatter plot of allele support
    """
    num_points = len(refs_11)+len(refs_10)+len(refs_00)+len(refs_nn)
    point_alpha = 10 / math.sqrt(max(100, num_points))
    plt.scatter(x=[ref+random.uniform(-0.5, 0.5) for ref in refs_11],
                y=[alt+random.uniform(-0.5, 0.5) for alt in alts_11],
                c='tab:red',
                alpha=point_alpha,
                label='1/1',
                edgecolors='none')
    plt.scatter(x=[ref+random.uniform(-0.5, 0.5) for ref in refs_10],
                y=[alt+random.uniform(-0.5, 0.5) for alt in alts_10],
                c='tab:purple',
                alpha=point_alpha,
                label='0/1',
                edgecolors='none')
    plt.scatter(x=[ref+random.uniform(-0.5, 0.5) for ref in refs_00],
                y=[alt+random.uniform(-0.5, 0.5) for alt in alts_00],
                c='tab:blue',
                alpha=point_alpha,
                label='0/0',
                edgecolors='none')
    plt.scatter(x=[ref+random.uniform(-0.5, 0.5) for ref in refs_nn],
                y=[alt+random.uniform(-0.5, 0.5) for alt in alts_nn],
                c='tab:brown',
                alpha=point_alpha,
                label='./.',
                edgecolors='none')
    axes = plt.gca()
    axes.set_xlim([0,60])
    axes.set_ylim([0,60])
    plt.xlabel('Reference allele support')
    plt.ylabel('Variant allele support')
    leg = plt.legend(frameon=True,
               fontsize="medium")
    for lh in leg.legendHandles:
        lh.set_alpha(1.0)

    plt.tight_layout()
    plt.savefig(output)
    plt.clf()


import sys
import os
import logging
import argparse


def parse_arguments(program_version, arguments = sys.argv[1:]):
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description="""SVIM (pronounced SWIM) is a structural variant caller for long reads. 
It discriminates six different variant classes: deletions, tandem and interspersed duplications, 
inversions, insertions and translocations. SVIM is unique in its capability of extracting both the genomic origin and 
destination of duplications.

SVIM consists of four major steps:
- COLLECT detects signatures for SVs in long read alignments
- CLUSTER merges signatures that come from the same SV
- COMBINE combines clusters from different genomic regions and classifies them into distinct SV types
- GENOTYPE uses alignments spanning SVs to determine their genotype

SVIM can process two types of input. Firstly, it can detect SVs from raw reads by aligning them to a given reference genome first ("SVIM.py reads [options] working_dir reads genome").
Alternatively, it can detect SVs from existing reads alignments in SAM/BAM format ("SVIM.py alignment [options] working_dir bam_file").
""")

    subparsers = parser.add_subparsers(help='modes', dest='sub')
    parser.add_argument('--version',
                        '-v',
                        action='version',
                        version='%(prog)s {version}'.format(version=program_version))
    parser_fasta = subparsers.add_parser('reads',
                                          help='Detect SVs from raw reads. Align reads to given reference genome first.')
    parser_fasta.add_argument('working_dir',
                               type=str,
                               help='Working and output directory. \
                                     Existing files in the directory are overwritten. \
                                     If the directory does not exist, it is created.')
    parser_fasta.add_argument('reads',
                               type=str,
                               help='Read file (FASTA, FASTQ, gzipped FASTA, gzipped FASTQ or file list). \
                                     The read file has to have one of the following supported file endings: \
                                     FASTA: .fa, .fasta, .FA, .fa.gz, .fa.gzip, .fasta.gz, .fasta.gzip \
                                     FASTQ: .fq, .fastq, .FQ, .fq.gz, .fq.gzip, .fastq.gz, .fastq.gzip \
                                     FILE LIST: .fa.fn, fq.fn')
    parser_fasta.add_argument('genome',
                               type=str,
                               help='Reference genome file (FASTA)')
    parser_fasta.add_argument('--verbose',
                              action='store_true',
                              help='Enable more verbose logging (default: %(default)s)')
    group_fasta_align = parser_fasta.add_argument_group('ALIGN')
    group_fasta_align.add_argument('--cores',
                                        type=int,
                                        default=1,
                                        help='CPU cores to use for the alignment (default: %(default)s)')
    group_fasta_align.add_argument('--aligner',
                                        type=str,
                                        default="ngmlr",
                                        choices=["ngmlr", "minimap2"],
                                        help='Tool for read alignment: ngmlr or minimap2 (default: %(default)s)')
    group_fasta_align.add_argument('--nanopore',
                                        action='store_true',
                                        help='Use Nanopore settings for read alignment (default: %(default)s)')
    group_fasta_collect = parser_fasta.add_argument_group('COLLECT')
    group_fasta_collect.add_argument('--min_mapq',
                                        type=int,
                                        default=20,
                                        help='Minimum mapping quality of reads to consider (default: %(default)s). \
                                              Reads with a lower mapping quality are ignored.')
    group_fasta_collect.add_argument('--min_sv_size',
                                        type=int,
                                        default=40,
                                        help='Minimum SV size to detect (default: %(default)s). \
                                              SVIM can potentially detect events of any size but is limited by the \
                                              signal-to-noise ratio in the input alignments. That means that more \
                                              accurate reads and alignments enable the detection of smaller events. \
                                              For current PacBio or Nanopore data, we would recommend a minimum size \
                                              of 40bp or larger.')
    group_fasta_collect.add_argument('--max_sv_size',
                                        type=int,
                                        default=100000,
                                        help='Maximum SV size to detect (default: %(default)s). \
                                              This parameter is used to distinguish long deletions (and inversions) from \
                                              translocations which cannot be distinguished from the alignment alone. \
                                              Split read segments mapping far apart on the reference could either \
                                              indicate a very long deletion (inversion) or a translocation breakpoint. \
                                              SVIM calls a translocation breakpoint if the mapping distance is larger \
                                              than this parameter and a deletion (or inversion) if it is smaller or equal.')
    group_fasta_collect.add_argument('--segment_gap_tolerance',
                                        type=int,
                                        default=10,
                                        help='Maximum tolerated gap between adjacent alignment segments (default: %(default)s). \
                                              This parameter applies to gaps on the reference and the read. Example: \
                                              Deletions are detected from two subsequent segments of a split read that are mapped \
                                              far apart from each other on the reference. The segment gap tolerance determines \
                                              the maximum tolerated length of the read gap between both segments. If there is an \
                                              unaligned read segment larger than this value between the two segments, no deletion is called.')
    group_fasta_collect.add_argument('--segment_overlap_tolerance',
                                        type=int,
                                        default=5,
                                        help='Maximum tolerated overlap between adjacent alignment segments (default: %(default)s). \
                                              This parameter applies to overlaps on the reference and the read. Example: \
                                              Deletions are detected from two subsequent segments of a split read that are mapped \
                                              far apart from each other on the reference. The segment overlap tolerance determines \
                                              the maximum tolerated length of an overlap between both segments on the read. If the \
                                              overlap between the two segments on the read is larger than this value, no deletion is called.')
    group_fasta_collect.add_argument('--all_bnds',
                                        action='store_true',
                                        help='Output all rearrangements additionally in BND notation (default: %(default)s). \
                                              By default, SV signatures from the read alignments are used to detect complete SVs, \
                                              such as deletions, insertions and inversions. When this option is enabled, all SVs \
                                              are also output in breakend (BND) notation as defined in the VCF specs. For instance, \
                                              a deletion gets two records in the VCF output: 1. the normal <DEL> record and 2. \
                                              a <BND> record representing the novel adjacency between the deletion\'s start and \
                                              end coordinate in the sample genome.')
    group_fasta_cluster = parser_fasta.add_argument_group('CLUSTER')
    group_fasta_cluster.add_argument('--partition_max_distance',
                                        type=int,
                                        default=1000,
                                        help='Maximum distance in bp between SVs in a partition (default: %(default)s). \
                                              Before clustering, the SV signatures are divided into coarse partitions. This parameter \
                                              determines the maximum distance between two subsequent signatures in the same partition. \
                                              If the distance between two subsequent signatures \
                                              is larger than this parameter, they are distributed into separate partitions.')
    group_fasta_cluster.add_argument('--position_distance_normalizer',
                                        type=int,
                                        default=900,
                                        help='Distance normalizer used for span-position distance (default: %(default)s). \
                                              SVIM clusters the SV signatures using an hierarchical clustering approach and a \
                                              novel distance metric called \"span-position distance\". Span-position distance \
                                              is the sum of two components, span distance and position distance. \
                                              The span distance is the difference in lengths between signatures normalized \
                                              by the greater length and always lies in the interval [0,1]. \
                                              The position distance is the difference in position between signatures \
                                              normalized by the distance normalizer (this parameter). For a position difference \
                                              of 1.8kb and a distance normalizer of 900, the position distance will be 2. \
                                              A smaller distance normalizer leads to a higher position distance and as a \
                                              consequence increases the importance of the position distance in the \
                                              span-position distance relative to the span distance.')
    group_fasta_cluster.add_argument('--edit_distance_normalizer',
                                      type=float,
                                      default=1.0,
                                      help='Distance normalizer used specifically for insertions (default: %(default)s). \
                                            SVIM clusters insertion signatures using an hierarchical clustering approach and a \
                                            special distance metric for insertions. This distance is the sum of two \
                                            components, position distance and edit distance between the insertion sequences. \
                                            The edit distance is normalized (i.e. divided) by the product of the span of the \
                                            longer insertion and this normalizer. The position distance is the difference in \
                                            position between signatures normalized by the position distance normalizer \
                                            (another parameter). A smaller edit distance normalizer leads to a larger \
                                            edit distance and as a consequence increases the importance of the edit distance \
                                            in the clustering process so that only insertions with very similar sequences \
                                            are clustered together. A larger edit distance normalizer diminishes the importance \
                                            of the insertion sequences in the clustering process.')
    group_fasta_cluster.add_argument('--cluster_max_distance',
                                        type=float,
                                        default=0.5,
                                        help='Maximum span-position distance between SVs in a cluster (default: %(default)s). \
                                              This is the most important parameter because it determines the strictness \
                                              of clustering. Choosing a large value leads to fewer but larger clusters with larger \
                                              distances between its members. Choosing a small value leads to more but smaller \
                                              clusters with smaller distances between its members. \
                                              This parameter determines the height of the cut-off in the hierarchical clustering \
                                              dendrogram.')
    group_fasta_combine = parser_fasta.add_argument_group('COMBINE')
    group_fasta_combine.add_argument('--del_ins_dup_max_distance',
                                        type=float,
                                        default=1.0,
                                        help='Maximum span-position distance between the origin of an insertion and a deletion to be flagged as a potential cut&paste insertion (default: %(default)s)')
    group_fasta_combine.add_argument('--trans_sv_max_distance',
                                        type=int,
                                        default=500,
                                        help='Maximum distance in bp between a translocation breakpoint and an SV signature to be combined (default: %(default)s)')
    group_fasta_combine.add_argument('--skip_consensus',
                                        action='store_true',
                                        help='Disable consensus computation for insertions (default: %(default)s). \
                                              This reduces the time and memory consumption of SVIM and might be \
                                              useful if consensus sequences are not needed. With this option, \
                                              insertion calls are represented by symbolic alleles (<INS>) instead \
                                              of sequence alles in the output VCF. \
                                              Consensus computation requires a modern CPU with the SSE 4.1 \
                                              instruction set. For older CPUs missing this instruction set, \
                                              consensus computation is automatically disabled.')
    group_fasta_combine.add_argument('--max_consensus_length',
                                        type=int,
                                        default=10000,
                                        help='Maximum size of insertion sequences for consensus computation. (default: %(default)s)\
                                              For insertions longer than this threshold, no consensus is computed to save memory.')
    group_fasta_genotype = parser_fasta.add_argument_group('GENOTYPE')
    group_fasta_genotype.add_argument('--skip_genotyping',
                                        action='store_true',
                                        help='Disable genotyping (default: %(default)s)')
    group_fasta_genotype.add_argument('--minimum_score',
                                        type=int,
                                        default=3,
                                        help='Minimum score for genotyping (default: %(default)s). \
                                              Only SV candidates with a higher or equal score are genotyped. Depending on the \
                                              score distribution among the SV candidates, decreasing this value increases the \
                                              runtime. We recommend to choose a value close to the score threshold used \
                                              for filtering the SV candidates.')
    group_fasta_genotype.add_argument('--homozygous_threshold',
                                        type=float,
                                        default=0.8,
                                        help='Minimum variant allele frequency to be called as homozygous (default: %(default)s). \
                                              Allele frequency is computed as the fraction of reads supporting the variant over the \
                                              total number of reads covering the variant. Variants with an allele frequence greater \
                                              than or equal to this threshold are called as homozygous alternative.')
    group_fasta_genotype.add_argument('--heterozygous_threshold',
                                        type=float,
                                        default=0.2,
                                        help='Minimum variant allele frequency to be called as heterozygous (default: %(default)s). \
                                              Allele frequency is computed as the fraction of reads supporting the variant over the \
                                              total number of reads covering the variant. Variants with an allele frequence greater \
                                              than or equal to this threshold but lower than the homozygous threshold are called as \
                                              heterozygous alternative. Variants with an allele frequence lower than this threshold \
                                              are called as homozygous reference.')
    group_fasta_genotype.add_argument('--minimum_depth',
                                        type=int,
                                        default=4,
                                        help='Minimum total read depth for genotyping (default: %(default)s). \
                                              Variants covered by a total number of reads lower than this value are not assigned \
                                              a genotype (./. in the output VCF file).')
    group_fasta_output = parser_fasta.add_argument_group('OUTPUT')
    group_fasta_output.add_argument('--sample',
                                        type=str,
                                        default="Sample",
                                        help='Sample ID to include in output vcf file (default: %(default)s)')
    group_fasta_output.add_argument('--types',
                                        type=str,
                                        default="DEL,INS,INV,DUP:TANDEM,DUP:INT,BND",
                                        help='SV types to include in output VCF (default: %(default)s). \
                                              Give a comma-separated list of SV types. The possible SV types are: DEL (deletions), \
                                              INS (novel insertions), INV (inversions), DUP:TANDEM (tandem duplications), \
                                              DUP:INT (interspersed duplications), BND (breakends).')
    group_fasta_output.add_argument('--symbolic_alleles',
                                        action='store_true',
                                        help='Use symbolic alleles, such as <DEL> or <INV> in output VCF (default: %(default)s). \
                                              By default, all SV alleles are represented by nucleotide sequences.')
    group_fasta_output.add_argument('--insertion_sequences',
                                        action='store_true',
                                        help='Output insertion sequences in INFO tag of VCF (default: %(default)s). \
                                              If enabled, the INFO/SEQS tag contains a list of insertion sequences from the supporting reads.')
    group_fasta_output.add_argument('--tandem_duplications_as_insertions',
                                        action='store_true',
                                        help='Represent tandem duplications as insertions in output VCF (default: %(default)s). \
                                              By default, tandem duplications are represented by the SVTYPE=DUP:TANDEM and the genomic source is given by the \
                                              POS and END tags. When enabling this option, duplications are instead represented by the SVTYPE=INS \
                                              and POS and END both give the insertion point of the duplication.')
    group_fasta_output.add_argument('--interspersed_duplications_as_insertions',
                                        action='store_true',
                                        help='Represent interspersed duplications as insertions in output VCF (default: %(default)s). \
                                              By default, interspersed duplications are represented by the SVTYPE=DUP:INT and the genomic source is given by the \
                                              POS and END tags. When enabling this option, duplications are instead represented by the SVTYPE=INS \
                                              and POS and END both give the insertion point of the duplication.')
    group_fasta_output.add_argument('--read_names',
                                        action='store_true',
                                        help='Output names of supporting reads in INFO tag of VCF (default: %(default)s). \
                                              If enabled, the INFO/READS tag contains the list of names of the supporting reads.')
    group_fasta_output.add_argument('--zmws',
                                        action='store_true',
                                        help='look for information on ZMWs in PacBio read names (default: %(default)s). \
                                              If enabled, the INFO/ZMWS tag contains the number of ZMWs that produced supporting reads.')

    parser_bam = subparsers.add_parser('alignment',
                                        help='Detect SVs from an existing alignment')
    parser_bam.add_argument('working_dir',
                             type=os.path.abspath,
                             help='Working and output directory. \
                                   Existing files in the directory are overwritten. \
                                   If the directory does not exist, it is created.')
    parser_bam.add_argument('bam_file',
                             type=str,
                             help='Coordinate-sorted and indexed BAM file with aligned long reads')
    parser_bam.add_argument('genome',
                               type=str,
                               help='Reference genome file that the long reads were aligned to (FASTA)')
    parser_bam.add_argument('--verbose',
                              action='store_true',
                              help='Enable more verbose logging (default: %(default)s)')
    group_bam_collect = parser_bam.add_argument_group('COLLECT')
    group_bam_collect.add_argument('--min_mapq',
                                      type=int,
                                      default=20,
                                      help='Minimum mapping quality of reads to consider (default: %(default)s). \
                                            Reads with a lower mapping quality are ignored.')
    group_bam_collect.add_argument('--min_sv_size',
                                      type=int,
                                      default=40,
                                      help='Minimum SV size to detect (default: %(default)s). \
                                            SVIM can potentially detect events of any size but is limited by the \
                                            signal-to-noise ratio in the input alignments. That means that more \
                                            accurate reads and alignments enable the detection of smaller events. \
                                            For current PacBio or Nanopore data, we would recommend a minimum size \
                                            of 40bp or larger.')
    group_bam_collect.add_argument('--max_sv_size',
                                      type=int,
                                      default=100000,
                                      help='Maximum SV size to detect (default: %(default)s). \
                                              This parameter is used to distinguish long deletions (and inversions) from \
                                              translocations which cannot be distinguished from the alignment alone. \
                                              Split read segments mapping far apart on the reference could either \
                                              indicate a very long deletion (inversion) or a translocation breakpoint. \
                                              SVIM calls a translocation breakpoint if the mapping distance is larger \
                                              than this parameter and a deletion (or inversion) if it is smaller or equal.')
    group_bam_collect.add_argument('--segment_gap_tolerance',
                                      type=int,
                                      default=10,
                                      help='Maximum tolerated gap between adjacent alignment segments (default: %(default)s). \
                                            This parameter applies to gaps on the reference and the read. Example: \
                                            Deletions are detected from two subsequent segments of a split read that are mapped \
                                            far apart from each other on the reference. The segment gap tolerance determines \
                                            the maximum tolerated length of the read gap between both segments. If there is an \
                                            unaligned read segment larger than this value between the two segments, no deletion is called.')
    group_bam_collect.add_argument('--segment_overlap_tolerance',
                                      type=int,
                                      default=5,
                                      help='Maximum tolerated overlap between adjacent alignment segments (default: %(default)s). \
                                            This parameter applies to overlaps on the reference and the read. Example: \
                                            Deletions are detected from two subsequent segments of a split read that are mapped \
                                            far apart from each other on the reference. The segment overlap tolerance determines \
                                            the maximum tolerated length of an overlap between both segments on the read. If the \
                                            overlap between the two segments on the read is larger than this value, no deletion is called.')
    group_bam_cluster = parser_bam.add_argument_group('CLUSTER')
    group_bam_cluster.add_argument('--partition_max_distance',
                                      type=int,
                                      default=1000,
                                      help='Maximum distance in bp between SVs in a partition (default: %(default)s). \
                                            Before clustering, the SV signatures are divided into coarse partitions. This parameter \
                                            determines the maximum distance between two subsequent signatures in the same partition. \
                                            If the distance between two subsequent signatures \
                                            is larger than this parameter, they are distributed into separate partitions.')
    group_bam_cluster.add_argument('--position_distance_normalizer',
                                      type=int,
                                      default=900,
                                      help='Distance normalizer used for span-position distance (default: %(default)s). \
                                            SVIM clusters the SV signatures using an hierarchical clustering approach and a \
                                            novel distance metric called \"span-position distance\". Span-position distance \
                                            is the sum of two components, span distance and position distance. \
                                            The span distance is the difference in lengths between signatures normalized \
                                            by the greater length and always lies in the interval [0,1]. \
                                            The position distance is the difference in position between signatures \
                                            normalized by the distance normalizer (this parameter). For a position difference \
                                            of 1.8kb and a distance normalizer of 900, the position distance will be 2. \
                                            A smaller distance normalizer leads to a higher position distance and as a \
                                            consequence increases the importance of the position distance in the \
                                            span-position distance relative to the span distance.')
    group_bam_cluster.add_argument('--edit_distance_normalizer',
                                      type=float,
                                      default=1.0,
                                      help='Distance normalizer used specifically for insertions (default: %(default)s). \
                                            SVIM clusters insertion signatures using an hierarchical clustering approach and a \
                                            special distance metric for insertions. This distance is the sum of two \
                                            components, position distance and edit distance between the insertion sequences. \
                                            The edit distance is normalized (i.e. divided) by the product of the span of the \
                                            longer insertion and this normalizer. The position distance is the difference in \
                                            position between signatures normalized by the position distance normalizer \
                                            (another parameter). A smaller edit distance normalizer leads to a larger \
                                            edit distance and as a consequence increases the importance of the edit distance \
                                            in the clustering process so that only insertions with very similar sequences \
                                            are clustered together. A larger edit distance normalizer diminishes the importance \
                                            of the insertion sequences in the clustering process.')
    group_bam_cluster.add_argument('--cluster_max_distance',
                                      type=float,
                                      default=0.5,
                                      help='Maximum span-position distance between SVs in a cluster (default: %(default)s). \
                                            This is the most important parameter because it determines the strictness \
                                            of clustering. Choosing a large value leads to fewer but larger clusters with larger \
                                            distances between its members. Choosing a small value leads to more but smaller \
                                            clusters with smaller distances between its members. \
                                            This parameter determines the height of the cut-off in the hierarchical clustering \
                                            dendrogram.')
    group_bam_cluster.add_argument('--all_bnds',
                                        action='store_true',
                                        help='Output all rearrangements additionally in BND notation (default: %(default)s). \
                                              By default, SV signatures from the read alignments are used to detect complete SVs, \
                                              such as deletions, insertions and inversions. When this option is enabled, all SVs \
                                              are also output in breakend (BND) notation as defined in the VCF specs. For instance, \
                                              a deletion gets two records in the VCF output: 1. the normal <DEL> record and 2. \
                                              a <BND> record representing the novel adjacency between the deletion\'s start and \
                                              end coordinate in the sample genome.')
    group_bam_combine = parser_bam.add_argument_group('COMBINE')
    group_bam_combine.add_argument('--del_ins_dup_max_distance',
                                      type=float,
                                      default=1.0,
                                      help='Maximum span-position distance between the origin of an insertion and a deletion to be flagged as a potential cut&paste insertion (default: %(default)s)')
    group_bam_combine.add_argument('--trans_sv_max_distance',
                                      type=int,
                                      default=500,
                                      help='Maximum distance in bp between a translocation breakpoint and an SV signature to be combined (default: %(default)s)')
    group_bam_combine.add_argument('--skip_consensus',
                                    action='store_true',
                                    help='Disable consensus computation for insertions (default: %(default)s). \
                                          This reduces the time and memory consumption of SVIM and might be \
                                          useful if consensus sequences are not needed. With this option, \
                                          insertion calls are represented by symbolic alleles (<INS>) instead \
                                          of sequence alles in the output VCF. \
                                          Consensus computation requires a modern CPU with the SSE 4.1 \
                                          instruction set. For older CPUs missing this instruction set, \
                                          consensus computation is automatically disabled.')
    group_bam_combine.add_argument('--max_consensus_length',
                                        type=int,
                                        default=10000,
                                        help='Maximum size of insertion sequences for consensus computation. (default: %(default)s)\
                                              For insertions longer than this threshold, no consensus is computed to save memory.')
    group_bam_genotype = parser_bam.add_argument_group('GENOTYPE')
    group_bam_genotype.add_argument('--skip_genotyping',
                                        action='store_true',
                                        help='Disable genotyping (default: %(default)s)')
    group_bam_genotype.add_argument('--minimum_score',
                                      type=int,
                                      default=3,
                                      help='Minimum score for genotyping (default: %(default)s). \
                                            Only SV candidates with a higher or equal score are genotyped. Depending on the \
                                            score distribution among the SV candidates, decreasing this value increases the \
                                            runtime. We recommend to choose a value close to the score threshold used \
                                            for filtering the SV candidates.')
    group_bam_genotype.add_argument('--homozygous_threshold',
                                      type=float,
                                      default=0.8,
                                      help='Minimum variant allele frequency to be called as homozygous (default: %(default)s). \
                                            Allele frequency is computed as the fraction of reads supporting the variant over the \
                                            total number of reads covering the variant. Variants with an allele frequence greater \
                                            than or equal to this threshold are called as homozygous alternative.')
    group_bam_genotype.add_argument('--heterozygous_threshold',
                                      type=float,
                                      default=0.2,
                                      help='Minimum variant allele frequency to be called as heterozygous (default: %(default)s). \
                                            Allele frequency is computed as the fraction of reads supporting the variant over the \
                                            total number of reads covering the variant. Variants with an allele frequence greater \
                                            than or equal to this threshold but lower than the homozygous threshold are called as \
                                            heterozygous alternative. Variants with an allele frequence lower than this threshold \
                                            are called as homozygous reference.')
    group_bam_genotype.add_argument('--minimum_depth',
                                      type=int,
                                      default=4,
                                      help='Minimum total read depth for genotyping (default: %(default)s). \
                                            Variants covered by a total number of reads lower than this value are not assigned \
                                            a genotype (./. in the output VCF file).')
    group_bam_output = parser_bam.add_argument_group('OUTPUT')
    group_bam_output.add_argument('--sample',
                                        type=str,
                                        default="Sample",
                                        help='Sample ID to include in output vcf file (default: %(default)s)')
    group_bam_output.add_argument('--types',
                                        type=str,
                                        default="DEL,INS,INV,DUP:TANDEM,DUP:INT,BND",
                                        help='SV types to include in output VCF (default: %(default)s). \
                                              Give a comma-separated list of SV types. The possible SV types are: DEL (deletions), \
                                              INS (novel insertions), INV (inversions), DUP:TANDEM (tandem duplications), \
                                              DUP:INT (interspersed duplications), BND (breakends).')
    group_bam_output.add_argument('--symbolic_alleles',
                                        action='store_true',
                                        help='Use symbolic alleles, such as <DEL> or <INV> in output VCF (default: %(default)s). \
                                              By default, all SV alleles are represented by nucleotide sequences.')
    group_bam_output.add_argument('--insertion_sequences',
                                        action='store_true',
                                        help='Output insertion sequences in INFO tag of VCF (default: %(default)s). \
                                              If enabled, the INFO/SEQS tag contains a list of insertion sequences from the supporting reads.')
    group_bam_output.add_argument('--tandem_duplications_as_insertions',
                                        action='store_true',
                                        help='Represent tandem duplications as insertions in output VCF (default: %(default)s). \
                                              By default, tandem duplications are represented by the SVTYPE=DUP:TANDEM and the genomic source is given by the \
                                              POS and END tags. When enabling this option, duplications are instead represented by the SVTYPE=INS \
                                              and POS and END both give the insertion point of the duplication.')
    group_bam_output.add_argument('--interspersed_duplications_as_insertions',
                                        action='store_true',
                                        help='Represent interspersed duplications as insertions in output VCF (default: %(default)s). \
                                              By default, interspersed duplications are represented by the SVTYPE=DUP:INT and the genomic source is given by the \
                                              POS and END tags. When enabling this option, duplications are instead represented by the SVTYPE=INS \
                                              and POS and END both give the insertion point of the duplication.')
    group_bam_output.add_argument('--read_names',
                                        action='store_true',
                                        help='Output names of supporting reads in INFO tag of VCF (default: %(default)s). \
                                              If enabled, the INFO/READS tag contains the list of names of the supporting reads.')
    group_bam_output.add_argument('--zmws',
                                        action='store_true',
                                        help='look for information on ZMWs in PacBio read names (default: %(default)s). \
                                              If enabled, the INFO/ZMWS tag contains the number of ZMWs that produced supporting reads.')

    return parser.parse_args(arguments)


def guess_file_type(reads_path):
    if reads_path.endswith(".fa") or reads_path.endswith(".fasta") or reads_path.endswith(".FA"):
        logging.info("Recognized reads file as FASTA format.")
        return "fasta"
    elif reads_path.endswith(".fq") or reads_path.endswith(".fastq") or reads_path.endswith(".FQ"):
        logging.info("Recognized reads file as FASTQ format.")
        return "fastq"
    elif reads_path.endswith(".fa.gz") or reads_path.endswith(".fasta.gz") or reads_path.endswith(".FA.gz") or reads_path.endswith(".fa.gzip") or reads_path.endswith(".fasta.gzip") or reads_path.endswith(".FA.gzip"):
        logging.info("Recognized reads file as gzipped FASTA format.")
        return "fasta_gzip"
    elif reads_path.endswith(".fq.gz") or reads_path.endswith(".fastq.gz") or reads_path.endswith(".FQ.gz") or reads_path.endswith(".fq.gzip") or reads_path.endswith(".fastq.gzip") or reads_path.endswith(".FQ.gzip"):
        logging.info("Recognized reads file as gzipped FASTQ format.")
        return "fastq_gzip"
    elif reads_path.endswith(".fa.fn") or reads_path.endswith(".fasta.fn") or reads_path.endswith(".FA.fn") or reads_path.endswith(".fq.fn") or reads_path.endswith(".fastq.fn") or reads_path.endswith(".FQ.fn"):
        logging.info("Recognized reads file as file list format.")
        return "list"
    else:
        logging.error("Unknown file ending of file {0}. See github.com/eldariont/svim/wiki/ for supported file endings. Exiting.".format(reads_path))
        return "unknown"


def read_file_list(path):
    file_list = open(path, "r")
    for line in file_list:
        yield line.strip()
    file_list.close()